import numpy
from . import transfers
from ..cosmology import Cosmology

class LinearPower(object):
    """
    An object to compute the linear power spectrum and related quantities,
    using a transfer function from the CLASS code or the analytic
    Eisenstein & Hu approximation.

    Parameters
    ----------
    cosmo : :class:`Cosmology`, astropy.cosmology.FLRW
        the cosmology instance; astropy cosmology objects are automatically
        converted to the proper type
    redshift : float
        the redshift of the power spectrum
    transfer : str, optional
        string specifying the transfer function to use; one of
        'CLASS', 'EisensteinHu', 'NoWiggleEisensteinHu'

    Attributes
    ----------
    cosmo : class:`Cosmology`
        the object giving the cosmological parameters
    sigma8 : float
        the z=0 amplitude of matter fluctuations
    redshift : float
        the redshift to compute the power at
    transfer : str
        the type of transfer function used
    """
    def __init__(self, cosmo, redshift, transfer='CLASS'):
        from astropy.cosmology import FLRW

        # convert astropy
        if isinstance(cosmo, FLRW):
            from nbodykit.cosmology import Cosmology
            cosmo = Cosmology.from_astropy(cosmo)

        # store a copy of the cosmology
        self.cosmo = cosmo.clone()

        # set sigma8 to the cosmology value
        self._sigma8 = self.cosmo.sigma8

        # setup the transfers
        if transfer not in transfers.available:
            raise ValueError("'transfer' should be one of %s" %str(transfers.available))
        self.transfer = transfer

        # initialize internal transfers
        c = self.cosmo.clone() # transfers get an internal copy
        self._transfer = getattr(transfers, transfer)(c, redshift)
        self._fallback = transfers.EisensteinHu(c, redshift) # fallback to analytic when out of range

        # normalize to proper sigma8
        self._norm = 1.
        self.redshift = 0;
        self._norm = (self._sigma8 / self.sigma_r(8.))**2 # sigma_r(z=0, r=8)

        # set redshift
        self.redshift = redshift

        # store meta-data
        self._attrs = {}
        self._attrs['transfer'] = transfer
        self._attrs['cosmo'] = dict(cosmo)

    @property
    def attrs(self):
        """
        The meta-data dictionary
        """
        self._attrs['redshift'] = self.redshift
        self._attrs['sigma8'] = self.sigma8
        return self._attrs

    @property
    def redshift(self):
        """
        The redshift of the power spectrum
        """
        return self._z

    @redshift.setter
    def redshift(self, value):
        self._z = value
        self._transfer.redshift = value
        self._fallback.redshift = value

    @property
    def sigma8(self):
        """
        The present day value of ``sigma_r(r=8 Mpc/h)``, used to normalize
        the power spectrum, which is proportional to the square of this value.

        The power spectrum can re-normalized by setting a different
        value for this parameter
        """
        return self._sigma8

    @sigma8.setter
    def sigma8(self, value):
        """
        Set the sigma8 value and normalize the power spectrum to the new value
        """
        # re-scale the normalization
        self._norm *= (value / self._sigma8)**2

        # update to this sigma8
        self._sigma8 = value

    def __call__(self, k):
        """
        Return the linear power spectrum in units of
        :math:`h^{-3} \mathrm{Mpc}^3` at the redshift specified by
        :attr:`redshift`.

        The transfer function used to evaluate the power spectrum is
        specified by the ``transfer`` attribute.

        Parameters
        ---------
        k : float, array_like
            the wavenumber in units of :math:`h Mpc^{-1}`

        Returns
        -------
        Pk : float, array_like
            the linear power spectrum evaluated at ``k`` in units of
            :math:`h^{-3} \mathrm{Mpc}^3`
        """
        if self.transfer != "CLASS":
            Pk = k**self.cosmo.n_s * self._transfer(k)**2
        else:
            k = numpy.asarray(k)
            kmax = self.cosmo.P_k_max
            inrange = k < 0.99999*kmax # prevents rounding errors

            # the return array (could be scalar array)
            Pk = numpy.zeros_like(k)

            # k values in and out of valid range
            k_in = k[inrange]; k_out = k[~inrange]

            # use CLASS in range
            Pk[inrange] = k_in**self.cosmo.n_s * self._transfer(k_in)**2

            # use Eisentein-Hu out of range
            if len(k_out):
                analytic_Tk = self._fallback(k_out)
                analytic_Tk *= self._transfer(kmax)/ self._fallback(kmax)
                Pk[~inrange] = k_out**self.cosmo.n_s * analytic_Tk**2

        return self._norm * Pk

    def velocity_dispersion(self, kmin=1e-5, kmax=10., **kwargs):
        r"""
        The velocity dispersion in units of of :math:`\mathrm{Mpc/h}` at
        ``redshift``.

        This returns :math:`\sigma_v`, defined as

        .. math::

            \sigma_v^2 = \frac{1}{3} \int_a^b \frac{d^3 q}{(2\pi)^3} \frac{P(q,z)}{q^2}.

        Parameters
        ----------
        kmin : float, optional
            the lower bound for the integral, in units of :math:`\mathrm{Mpc/h}`
        kmax : float, optional
            the upper bound for the integral, in units of :math:`\mathrm{Mpc/h}`
        """
        from scipy.integrate import quad

        def integrand(logq):
            q = numpy.exp(logq)
            return q*self(q)
        sigmasq = quad(integrand, numpy.log(kmin), numpy.log(kmax), **kwargs)[0] / (6*numpy.pi**2)
        return sigmasq**0.5

    def sigma_r(self, r, kmin=1e-5, kmax=1e1):
        r"""
        The mass fluctuation within a sphere of radius ``r``, in
        units of :math:`h^{-1} Mpc` at ``redshift``.

        This returns :math:`\sigma`, where

        .. math::

            \sigma^2 = \int_0^\infty \frac{k^3 P(k,z)}{2\pi^2} W^2_T(kr) \frac{dk}{k},

        where :math:`W_T(x) = 3/x^3 (\mathrm{sin}x - x\mathrm{cos}x)` is
        a top-hat filter in Fourier space.

        The value of this function with ``r=8`` returns
        :attr:`sigma8`, within numerical precision.

        Parameters
        ----------
        r : float, array_like
            the scale to compute the mass fluctation over, in units of
            :math:`h^{-1} Mpc`
        kmin : float, optional
            the lower bound for the integral, in units of :math:`\mathrm{Mpc/h}`
        kmax : float, optional
            the upper bound for the integral, in units of :math:`\mathrm{Mpc/h}`
        """
        import mcfit
        from scipy.interpolate import InterpolatedUnivariateSpline as spline

        k = numpy.logspace(numpy.log10(kmin), numpy.log10(kmax), 1024)
        Pk = self(k)
        R, sigmasq = mcfit.TophatVar(k)(Pk)

        return spline(R, sigmasq)(r)**0.5


def NoWiggleEHPower(cosmo, redshift):
    import warnings
    warnings.warn("NoWiggleEHPower is deprecated. Use LinearPower with transfer set to 'NoWiggleEisensteinHu'", FutureWarning)
    return LinearPower(cosmo=cosmo, redshift=redshift, transfer="NoWiggleEisensteinHu")

def EHPower(cosmo, redshift):
    import warnings
    warnings.warn("NoWiggleEHPower is deprecated. Use LinearPower with transfer set to 'EisensteinHu'", FutureWarning)
    return LinearPower(cosmo=cosmo, redshift=redshift, transfer="EisensteinHu")



from runtests.mpi import MPITest
from nbodykit.io.hdf import HDFFile
import os
import numpy
import tempfile
import pickle
import contextlib
import pytest

# h5py is an optional dependency
try: import h5py
except ImportError: h5py = None

@contextlib.contextmanager
def temporary_data():
    """
    Write some temporary HDF5 data to disk
    """
    try:
        # generate data
        dset = numpy.empty(1024, dtype=[('Position', ('f8', 3)), ('Mass', 'f8')])
        dset['Position'] = numpy.random.random(size=(1024, 3))
        dset['Mass'] = numpy.random.random(size=1024)

        # write to file
        tmpfile = tempfile.mkstemp()[1]
        with h5py.File(tmpfile , 'w') as ff:
            ff.create_dataset('X', data=dset) # store structured array as dataset
            grp = ff.create_group('Y')
            grp.create_dataset('Position', data=dset['Position']) # column as dataset
            grp.create_dataset('Mass', data=dset['Mass']) # column as dataset
            
        yield (dset, tmpfile)
    except:
        raise
    finally:
        os.unlink(tmpfile)
        
        
@MPITest([1])
@pytest.mark.skipif(h5py is None, "h5py is not installed")
def test_data(comm):

    with temporary_data() as (data, tmpfile):

        # read
        f = HDFFile(tmpfile)

        # check columns
        cols = ['X/Mass', 'X/Position', 'Y/Mass', 'Y/Position']
        assert(all(col in cols for col in f.columns))
        
        # make sure data is the same
        for col in cols:
            field = col.rsplit('/', 1)[-1]
            numpy.testing.assert_almost_equal(data[field], f[col][:])
            
        # try a slice
        data2 = f.read(cols, 0, 512, 1)
        for col in cols:
            field = col.rsplit('/', 1)[-1]
            numpy.testing.assert_almost_equal(data[field][:512], data2[col])

        # check size
        numpy.testing.assert_equal(f.size, 1024)


@MPITest([1])
@pytest.mark.skipif(h5py is None, "h5py is not installed")
def test_nonzero_root(comm):
    
    with temporary_data() as (data, tmpfile):
        
        # read
        f = HDFFile(tmpfile, root='Y')
        
        # non-zero root
        f = HDFFile(tmpfile, root='Y')
        assert(all(col in ['Mass', 'Position'] for col in f.columns))
        
        # wrong root
        with pytest.raises(ValueError):
            f = HDFFile(tmpfile, root='Z')
    

@MPITest([1])
@pytest.mark.skipif(h5py is None, "h5py is not installed")
def test_nonzero_exclude(comm):
    
    with temporary_data() as (data, tmpfile):
        
        # read
        f = HDFFile(tmpfile, exclude=['Y'])
        
        # non-zero exclude
        f = HDFFile(tmpfile, exclude=['Y'])
        assert(all(col in ['Mass', 'Position'] for col in f.columns))
        
        # non-zero exclude and root
        f = HDFFile(tmpfile, exclude=['Mass'], root='Y')
        assert all(col in ['Position'] for col in f.columns)
        
        # bad exclude
        with pytest.raises(ValueError):
            f = HDFFile(tmpfile, exclude=['Z'])
        

@MPITest([1])
@pytest.mark.skipif(h5py is None, "h5py is not installed")
def test_data_mismatch(comm):
    
    # generate data
    pos = numpy.random.random(size=(1024, 3))
    mass = numpy.random.random(size=512)
    
    # write to file
    tmpfile = tempfile.mkstemp()[1]
    with h5py.File(tmpfile , 'w') as ff:
        ff.create_dataset('Mass', data=mass) 
        ff.create_dataset('Position', data=pos)
    
    # fails due to mismatched sizes
    with pytest.raises(ValueError):
        f = HDFFile(tmpfile)
    
    # only one dataset now, so this works
    f = HDFFile(tmpfile, exclude=['Mass'])
    assert f.size == 1024

    os.unlink(tmpfile) 
     
   
@MPITest([1])
@pytest.mark.skipif(h5py is None, "h5py is not installed")
def test_empty(comm):
    
    # create empty file
    tmpfile = tempfile.mkstemp()[1]
    with h5py.File(tmpfile , 'w') as ff:
        ff.create_group('Y') 
    
    # no datasets!
    with pytest.raises(ValueError):
        f = HDFFile(tmpfile)
    
    os.unlink(tmpfile) 
        


from nbodykit.base.mesh import MeshSource
from nbodykit import CurrentMPIComm, mockmaker
from nbodykit.utils import attrs_to_dict
from pmesh.pm import RealField, ComplexField

class FieldMesh(MeshSource):
    """
    A MeshSource initialized from an in-memory Field object, either a
    :class:`pmesh.pm.RealField` or :class:`pmesh.pm.ComplexField`.

    .. note::
        The original field object is never modified by this source.

    Parameters
    ----------
    field : :class:`~pmesh.pm.RealField`, :class:`~pmesh.pm.ComplexField`
        the :mod:`pmesh` Field object, either of real or complex type
    Nmesh : int, 3-vector of int, optional
        the desired number of cells per size on the mesh. If this is different
        than the ``Nmesh`` of the input Field, the Field will be re-sampled
    """
    def __repr__(self):
        return "FieldMesh()"

    def __init__(self, field):

        MeshSource.__init__(self, field.pm.comm, field.Nmesh, field.BoxSize, field.pm.dtype)
        self.field = field

    def to_complex_field(self):
        """
        Return a copy of the (possibly re-sampled) input ComplexField
        """
        if isinstance(self.field, ComplexField):
            return self.field.copy()
        else:
            return NotImplemented

    def to_real_field(self):
        """
        Return a copy of the (possibly re-sampled) input RealField
        """
        if isinstance(self.field, RealField):
            return self.field.copy()
        else:
            return NotImplemented



import os
import numpy
import logging

from nbodykit import CurrentMPIComm
from nbodykit.binned_statistic import BinnedStatistic
from nbodykit.meshtools import SlabIterator
from nbodykit.base.catalog import CatalogSourceBase
from nbodykit.base.mesh import MeshSource
from nbodykit.base.catalogmesh import CatalogMesh

class FFTBase(object):
    """
    Base class provides functions for periodic FFT based Power spectrum code.

    Parameters
    ----------
    first : CatalogSource
        the first catalog source
    second : CatalogSource, None
        the second source, or None for auto-correlations
    Nmesh : int, 3-vector
        the number of cells per mesh size
    BoxSize : 3-vector
        the size of the box
    """
    def __init__(self, first, second, Nmesh, BoxSize):
        from pmesh.pm import ParticleMesh

        first = _cast_source(first, Nmesh=Nmesh, BoxSize=BoxSize)
        if second is not None:
            second = _cast_source(second, Nmesh=Nmesh, BoxSize=BoxSize)
        else:
            second = first

        self.first = first
        self.second = second

        # grab comm from first source
        self.comm = first.comm

        # check for comm mismatch
        assert second.comm is first.comm, "communicator mismatch between input sources"

        # check box size
        if not numpy.array_equal(first.attrs['BoxSize'], second.attrs['BoxSize']):
            raise ValueError("'BoxSize' mismatch between sources in FFTPower")

        # save meta-data
        self.attrs = {}
        self.attrs['Nmesh'] = first.attrs['Nmesh'].copy()
        self.attrs['BoxSize'] = first.attrs['BoxSize'].copy()

        self.attrs.update(zip(['Lx', 'Ly', 'Lz'], self.attrs['BoxSize']))
        self.attrs.update({'volume':self.attrs['BoxSize'].prod()})

    def save(self, output):
        """
        Save the result to disk. The format is currently JSON.
        """
        import json
        from nbodykit.utils import JSONEncoder

        # only the master rank writes
        if self.comm.rank == 0:
            self.logger.info('measurement done; saving result to %s' % output)

            with open(output, 'w') as ff:
                json.dump(self.__getstate__(), ff, cls=JSONEncoder)

    @classmethod
    @CurrentMPIComm.enable
    def load(cls, output, comm=None):
        """
        Load a saved result. The result has been saved to disk with :func:`save`.
        """
        import json
        from nbodykit.utils import JSONDecoder
        if comm.rank == 0:
            with open(output, 'r') as ff:
                state = json.load(ff, cls=JSONDecoder)
        else:
            state = None
        state = comm.bcast(state)
        self = object.__new__(cls)
        self.__setstate__(state)
        self.comm = comm

        return self

    def _compute_3d_power(self):
        """
        Compute and return the 3D power from two input sources

        Returns
        -------
        p3d : array_like (complex)
            the 3D complex array holding the power spectrum
        """
        c1 = self.first.paint(mode='complex', Nmesh=self.attrs['Nmesh'])

        # compute the auto power of single supplied field
        if self.first is self.second:
            c2 = c1
        else:
            c2 = self.second.paint(mode='complex', Nmesh=self.attrs['Nmesh'])

        # calculate the 3d power spectrum, slab-by-slab to save memory
        p3d = c1
        for (s0, s1, s2) in zip(p3d.slabs, c1.slabs, c2.slabs):
            s0[...] = s1 * s2.conj()

        for i, s0 in zip(p3d.slabs.i, p3d.slabs):
            # clear the zero mode.
            mask = True
            for i1 in i:
                mask = mask & (i1 == 0)
            s0[mask] = 0

        # the complex field is dimensionless; power is L^3
        # ref to http://icc.dur.ac.uk/~tt/Lectures/UA/L4/cosmology.pdf
        p3d[...] *= self.attrs['BoxSize'].prod()

        # get the number of objects (in a safe manner)
        N1 = c1.attrs.get('N', 0)
        N2 = c2.attrs.get('N', 0)
        self.attrs.update({'N1':N1, 'N2':N2})

        # add shotnoise (nonzero only for auto-spectra)
        Pshot = 0
        if self.first is self.second:
            if 'shotnoise' not in c1.attrs:
                if isinstance(self.first, CatalogMesh):
                    import warnings
                    warnings.warn(("no 'shotnoise' found for auto power spectrum "
                                   "of discrete data in FFTPower"))
            else:
                Pshot = c1.attrs['shotnoise']
        self.attrs['shotnoise'] = Pshot

        return p3d

class FFTPower(FFTBase):
    """
    Algorithm to compute the 1d or 2d power spectrum and/or multipoles
    in a periodic box, using a Fast Fourier Transform (FFT).

    This computes the power spectrum as the square of the Fourier modes of the
    density field, which are computed via a FFT.

    Results are computed when the object is inititalized. See the documenation
    of :func:`~FFTPower.run` for the attributes storing the results.

    .. note::
        A full tutorial on the class is available in the documentation
        :ref:`here <fftpower>`.

    Parameters
    ----------
    first : CatalogSource, MeshSource
        the source for the first field; if a CatalogSource is provided, it
        is automatically converted to MeshSource using the default painting
        parameters (via :func:`~nbodykit.base.catalogmesh.CatalogMesh.to_mesh`)
    mode : {'1d', '2d'}
        compute either 1d or 2d power spectra
    Nmesh : int, optional
        the number of cells per side in the particle mesh used to paint the source
    BoxSize : int, 3-vector, optional
        the size of the box
    second : CatalogSource, MeshSource, optional
        the second source for cross-correlations
    los : array_like , optional
        the direction to use as the line-of-sight; must be a unit vector
    Nmu : int, optional
        the number of mu bins to use from :math:`\mu=[0,1]`;
        if `mode = 1d`, then ``Nmu`` is set to 1
    dk : float, optional
        the linear spacing of ``k`` bins to use; if not provided, the
        fundamental mode  of the box is used
    kmin : float, optional
        the lower edge of the first ``k`` bin to use
    poles : list of int, optional
        a list of multipole numbers ``ell`` to compute :math:`P_\ell(k)`
        from :math:`P(k,\mu)`
    """
    logger = logging.getLogger('FFTPower')

    def __init__(self, first, mode, Nmesh=None, BoxSize=None, second=None,
                    los=[0, 0, 1], Nmu=5, dk=None, kmin=0., poles=[]):

        # mode is either '1d' or '2d'
        if mode not in ['1d', '2d']:
            raise ValueError("`mode` should be either '1d' or '2d'")

        if poles is None:
            poles = []

        # check los
        if numpy.isscalar(los) or len(los) != 3:
            raise ValueError("line-of-sight ``los`` should be vector with length 3")
        if not numpy.allclose(numpy.einsum('i,i', los, los), 1.0, rtol=1e-5):
            raise ValueError("line-of-sight ``los`` must be a unit vector")

        FFTBase.__init__(self, first, second, Nmesh, BoxSize)

        # save meta-data
        self.attrs['mode'] = mode
        self.attrs['los'] = los
        self.attrs['Nmu'] = Nmu
        self.attrs['poles'] = poles

        if dk is None:
            dk = 2 * numpy.pi / self.attrs['BoxSize'].min()

        self.attrs['dk'] = dk
        self.attrs['kmin'] = kmin

        self.run()

    def run(self):
        """
        Compute the power spectrum in a periodic box, using FFTs. This
        function returns nothing, but attaches several attributes
        to the class:

        - :attr:`edges`
        - :attr:`power`
        - :attr:`poles`

        Attributes
        ----------
        edges : array_like
            the edges of the wavenumber bins
        power : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            a BinnedStatistic object that holds the measured :math:`P(k)` or
            :math:`P(k,\mu)`. It stores the following variables:

            - k :
                the mean value for each ``k`` bin
            - mu : ``mode=2d`` only
                the mean value for each ``mu`` bin
            - power :
                complex array storing the real and imaginary components of the power
            - modes :
                the number of Fourier modes averaged together in each bin

        poles : :class:`~nbodykit.binned_statistic.BinnedStatistic` or ``None``
            a BinnedStatistic object to hold the multipole results
            :math:`P_\ell(k)`; if no multipoles were requested by the user,
            this is ``None``. It stores the following variables:

            - k :
                the mean value for each ``k`` bin
            - power_L :
                complex array storing the real and imaginary components for
                the :math:`\ell=L` multipole
            - modes :
                the number of Fourier modes averaged together in each bin

        attrs : dict
            dictionary of meta-data; in addition to storing the input parameters,
            it includes the following fields computed during the algorithm
            execution:

            - shotnoise : float
                the power Poisson shot noise, equal to :math:`V/N`, where
                :math:`V` is the volume of the box and `N` is the total
                number of objects; if a cross-correlation is computed, this
                will be equal to zero
            - N1 : int
                the total number of objects in the first source
            - N2 : int
                the total number of objects in the second source
        """

        # only need one mu bin if 1d case is requested
        if self.attrs['mode'] == "1d": self.attrs['Nmu'] = 1

        # measure the 3D power (y3d is a ComplexField)
        y3d = self._compute_3d_power()

        # binning in k out to the minimum nyquist frequency
        # (accounting for possibly anisotropic box)
        dk = self.attrs['dk']
        kmin = self.attrs['kmin']
        kedges = numpy.arange(kmin, numpy.pi*y3d.Nmesh.min()/y3d.BoxSize.max() + dk/2, dk)

        # project on to the desired basis
        muedges = numpy.linspace(0, 1, self.attrs['Nmu']+1, endpoint=True)
        edges = [kedges, muedges]
        result, pole_result = project_to_basis(y3d, edges,
                                               poles=self.attrs['poles'],
                                               los=self.attrs['los'])

        # format the power results into structured array
        if self.attrs['mode'] == "1d":
            cols = ['k', 'power', 'modes']
            icols = [0, 2, 3]
            edges = edges[0]
        else:
            cols = ['k', 'mu', 'power', 'modes']
            icols = [0, 1, 2, 3]

        # power results as a structured array
        dtype = numpy.dtype([(name, result[icol].dtype.str) for icol,name in zip(icols,cols)])
        power = numpy.squeeze(numpy.empty(result[0].shape, dtype=dtype))
        for icol, col in zip(icols, cols):
            power[col][:] = numpy.squeeze(result[icol])

        # multipole results as a structured array
        poles = None
        if pole_result is not None:
            k, poles, N = pole_result
            cols = ['k'] + ['power_%d' %l for l in self.attrs['poles']] + ['modes']
            result = [k] + [pole for pole in poles] + [N]

            dtype = numpy.dtype([(name, result[icol].dtype.str) for icol,name in enumerate(cols)])
            poles = numpy.empty(result[0].shape, dtype=dtype)
            for icol, col in enumerate(cols):
                poles[col][:] = result[icol]

        # set all the necessary results
        self.edges = edges
        self.poles = poles
        self.power = power

        self._make_datasets()

    def __getstate__(self):
        state = dict(
                     edges=self.edges,
                     power=self.power.data,
                     poles=getattr(self.poles, 'data', None),
                     attrs=self.attrs)
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._make_datasets()

    def _make_datasets(self):

        if self.attrs['mode'] == '1d':
            self.power = BinnedStatistic(['k'], [self.edges], self.power, fields_to_sum=['modes'], **self.attrs)
        else:
            self.power = BinnedStatistic(['k', 'mu'], self.edges, self.power, fields_to_sum=['modes'], **self.attrs)
        if self.poles is not None:
            self.poles = BinnedStatistic(['k'], [self.power.edges['k']], self.poles, fields_to_sum=['modes'], **self.attrs)

    def _compute_3d_power(self):
        """
        Compute and return the 3D power from two input sources

        Returns
        -------
        p3d : array_like (complex)
            the 3D complex array holding the power spectrum
        """
        c1 = self.first.paint(mode='complex', Nmesh=self.attrs['Nmesh'])

        # compute the auto power of single supplied field
        if self.first is self.second:
            c2 = c1
        else:
            c2 = self.second.paint(mode='complex', Nmesh=self.attrs['Nmesh'])

        # calculate the 3d power spectrum, slab-by-slab to save memory
        p3d = c1
        for (s0, s1, s2) in zip(p3d.slabs, c1.slabs, c2.slabs):
            s0[...] = s1 * s2.conj()

        for i, s0 in zip(p3d.slabs.i, p3d.slabs):
            # clear the zero mode.
            mask = True
            for i1 in i:
                mask = mask & (i1 == 0)
            s0[mask] = 0

        # the complex field is dimensionless; power is L^3
        # ref to http://icc.dur.ac.uk/~tt/Lectures/UA/L4/cosmology.pdf
        p3d[...] *= self.attrs['BoxSize'].prod()

        # get the number of objects (in a safe manner)
        N1 = c1.attrs.get('N', 0)
        N2 = c2.attrs.get('N', 0)
        self.attrs.update({'N1':N1, 'N2':N2})

        # add shotnoise (nonzero only for auto-spectra)
        Pshot = 0
        if self.first is self.second:
            if 'shotnoise' not in c1.attrs:
                if isinstance(self.first, CatalogMesh):
                    import warnings
                    warnings.warn(("no 'shotnoise' found for auto power spectrum "
                                   "of discrete data in FFTPower"))
            else:
                Pshot = c1.attrs['shotnoise']
        self.attrs['shotnoise'] = Pshot

        return p3d

class ProjectedFFTPower(FFTBase):
    """
    The power spectrum of a field in a periodic box, projected over certain axes.

    This is not really always physically meaningful, but convenient for
    making sense of Lyman-Alpha forest or lensing maps.

    This is usually called the 1d power spectrum or 2d power spectrum.

    Results are computed when the object is inititalized. See the documenation
    of :func:`~ProjectedFFTPower.run` for the attributes storing the results.

    Parameters
    ----------
    first : CatalogSource, MeshSource
        the source for the first field; if a CatalogSource is provided, it
        is automatically converted to MeshSource using the default painting
        parameters (via :func:`~nbodykit.base.catalogmesh.CatalogMesh.to_mesh`)
    Nmesh : int, optional
        the number of cells per side in the particle mesh used to paint the source
    BoxSize : int, 3-vector, optional
        the size of the box
    second : CatalogSource, MeshSource, optional
        the second source for cross-correlations
    axes : tuple
        axes to measure the power on. The axes not in the list will be averaged out.
        For example:
        - (0, 1) : project to x,y and measure power
        - (0) : project to x and measure power.
    dk : float, optional
        the linear spacing of ``k`` bins to use; if not provided, the
        fundamental mode  of the box is used
    kmin : float, optional
        the lower edge of the first ``k`` bin to use
    """
    logger = logging.getLogger('ProjectedFFTPower')

    def __init__(self, first, Nmesh=None, BoxSize=None, second=None,
                    axes=(0, 1), dk=None, kmin=0.):

        FFTBase.__init__(self, first, second, Nmesh, BoxSize)

        # only deal with 1d and 2d projections.
        assert len(axes) in (1, 2), "length of ``axes`` in ProjectedFFTPower should be 1 or 2"

        if dk is None:
            dk = 2 * numpy.pi / self.attrs['BoxSize'].min()

        self.attrs['dk'] = dk
        self.attrs['kmin'] = kmin

        self.attrs['axes'] = axes
        self.run()

    def run(self):
        """
        Run the algorithm. This attaches the following attributes to the class:

        - :attr:`edges`
        - :attr:`power`

        Attributes
        ----------
        edges : array_like
            the edges of the wavenumber bins
        power : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            a BinnedStatistic object that holds the projected power.
            It stores the following variables:

            - k :
                the mean value for each ``k`` bin
            - power :
                complex array holding the real and imaginary components of the
                projected power
            - modes :
                the number of Fourier modes averaged together in each bin
        """
        c1 = self.first.paint(Nmesh=self.attrs['Nmesh'], mode='complex')
        r1 = c1.preview(self.attrs['Nmesh'], axes=self.attrs['axes'])
        # average along projected axes;
        # part of product is the rfftn vs r2c (for axes)
        # the rest is for the mean (Nmesh - axes)
        c1 = numpy.fft.rfftn(r1) / self.attrs['Nmesh'].prod()

        # compute the auto power of single supplied field
        if self.first is self.second:
            c2 = c1
        else:
            c2 = self.second.paint(Nmesh=self.attrs['Nmesh'], mode='complex')
            r2 = c2.preview(self.attrs['Nmesh'], axes=self.attrs['axes'])
            c2 = numpy.fft.rfftn(r2) / self.attrs['Nmesh'].prod() # average along projected axes

        pk = c1 * c2.conj()
        # clear the zero mode
        pk.flat[0] = 0

        shape = numpy.array([self.attrs['Nmesh'][i] for i in self.attrs['axes']], dtype='int')
        boxsize = numpy.array([self.attrs['BoxSize'][i] for i in self.attrs['axes']])
        I = numpy.eye(len(shape), dtype='int') * -2 + 1

        k = [numpy.fft.fftfreq(N, 1. / (N * 2 * numpy.pi / L))[:pkshape].reshape(kshape) for N, L, kshape, pkshape in zip(shape, boxsize, I, pk.shape)]

        kmag = sum(ki ** 2 for ki in k) ** 0.5
        W = numpy.empty(pk.shape, dtype='f4')
        W[...] = 2.0
        W[..., 0] = 1.0
        W[..., -1] = 1.0

        dk = self.attrs['dk']
        kmin = self.attrs['kmin']
        axes = list(self.attrs['axes'])
        kedges = numpy.arange(kmin, numpy.pi * self.attrs['Nmesh'][axes].min() / self.attrs['BoxSize'][axes].max() + dk/2, dk)

        xsum = numpy.zeros(len(kedges) + 1)
        Psum = numpy.zeros(len(kedges) + 1, dtype='complex128')
        Nsum = numpy.zeros(len(kedges) + 1)

        dig = numpy.digitize(kmag.flat, kedges)
        xsum.flat += numpy.bincount(dig, weights=(W * kmag).flat, minlength=xsum.size)
        Psum.real.flat += numpy.bincount(dig, weights=(W * pk.real).flat, minlength=xsum.size)
        Psum.imag.flat += numpy.bincount(dig, weights=(W * pk.imag).flat, minlength=xsum.size)
        Nsum.flat += numpy.bincount(dig, weights=W.flat, minlength=xsum.size)

        self.power = numpy.empty(len(kedges) - 1,
                dtype=[('k', 'f8'), ('power', 'c16'), ('modes', 'f8')])

        with numpy.errstate(invalid='ignore'):
            self.power['k'] = (xsum / Nsum)[1:-1]
            self.power['power'] = (Psum / Nsum)[1:-1] * boxsize.prod() # dimension is 'volume'
            self.power['modes'] = Nsum[1:-1]

        self.edges = kedges

        self.power = BinnedStatistic(['k'], [self.edges], self.power)

    def __getstate__(self):
        state = dict(
                     edges=self.edges,
                     power=self.power.data,
                     attrs=self.attrs)
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.power = BinnedStatistic(['k'], [self.edges], self.power)

def project_to_basis(y3d, edges, los=[0, 0, 1], poles=[]):
    """
    Project a 3D statistic on to the specified basis. The basis will be one
    of:

    - 2D (`x`, `mu`) bins: `mu` is the cosine of the angle to the line-of-sight
    - 2D (`x`, `ell`) bins: `ell` is the multipole number, which specifies
      the Legendre polynomial when weighting different `mu` bins

    .. note::

        The 2D (`x`, `mu`) bins will be computed only if `poles` is specified.
        See return types for further details.

    Notes
    -----
    *   the `mu` range extends from 0.0 to 1.0
    *   the `mu` bins are half-inclusive half-exclusive, except the last bin
        is inclusive on both ends (to include `mu = 1.0`)

    Parameters
    ----------
    y3d : RealField or ComplexField
        the 3D array holding the statistic to be projected to the specified basis
    edges : list of arrays, (2,)
        list of arrays specifying the edges of the desired `x` bins and `mu` bins
    los : array_like,
        the line-of-sight direction to use, which `mu` is defined with
        respect to; default is [0, 0, 1] for z.
    poles : list of int, optional
        if provided, a list of integers specifying multipole numbers to
        project the 2d `(x, mu)` bins on to
    hermitian_symmetric : bool, optional
        Whether the input array `y3d` is Hermitian-symmetric, i.e., the negative
        frequency terms are just the complex conjugates of the corresponding
        positive-frequency terms; if ``True``, the positive frequency terms
        will be explicitly double-counted to account for this symmetry

    Returns
    -------
    result : tuple
        the 2D binned results; a tuple of ``(xmean_2d, mumean_2d, y2d, N_2d)``, where:

        - xmean_2d : array_like, (Nx, Nmu)
            the mean `x` value in each 2D bin
        - mumean_2d : array_like, (Nx, Nmu)
            the mean `mu` value in each 2D bin
        - y2d : array_like, (Nx, Nmu)
            the mean `y3d` value in each 2D bin
        - N_2d : array_like, (Nx, Nmu)
            the number of values averaged in each 2D bin

    pole_result : tuple or `None`
        the multipole results; if `poles` supplied it is a tuple of ``(xmean_1d, poles, N_1d)``,
        where:

        - xmean_1d : array_like, (Nx,)
            the mean `x` value in each 1D multipole bin
        - poles : array_like, (Nell, Nx)
            the mean multipoles value in each 1D bin
        - N_1d : array_like, (Nx,)
            the number of values averaged in each 1D bin
    """
    comm = y3d.pm.comm
    x3d = y3d.x
    hermitian_symmetric = numpy.iscomplexobj(y3d)

    from scipy.special import legendre

    # setup the bin edges and number of bins
    xedges, muedges = edges
    x2edges = xedges**2
    Nx = len(xedges) - 1
    Nmu = len(muedges) - 1

    # always make sure first ell value is monopole, which
    # is just (x, mu) projection since legendre of ell=0 is 1
    do_poles = len(poles) > 0
    _poles = [0]+sorted(poles) if 0 not in poles else sorted(poles)
    legpoly = [legendre(l) for l in _poles]
    ell_idx = [_poles.index(l) for l in poles]
    Nell = len(_poles)

    # valid ell values
    if any(ell < 0 for ell in _poles):
        raise ValueError("in `project_to_basis`, multipole numbers must be non-negative integers")

    # initialize the binning arrays
    musum = numpy.zeros((Nx+2, Nmu+2))
    xsum = numpy.zeros((Nx+2, Nmu+2))
    ysum = numpy.zeros((Nell, Nx+2, Nmu+2), dtype=y3d.dtype) # extra dimension for multipoles
    Nsum = numpy.zeros((Nx+2, Nmu+2), dtype='i8')

    # if input array is Hermitian symmetric, only half of the last
    # axis is stored in `y3d`
    symmetry_axis = -1 if hermitian_symmetric else None

    # iterate over y-z planes of the coordinate mesh
    for slab in SlabIterator(x3d, axis=0, symmetry_axis=symmetry_axis):

        # the square of coordinate mesh norm
        # (either Fourier space k or configuraton space x)
        xslab = slab.norm2()

        # if empty, do nothing
        if len(xslab.flat) == 0: continue

        # get the bin indices for x on the slab
        dig_x = numpy.digitize(xslab.flat, x2edges)

        # make xslab just x
        xslab **= 0.5

        # get the bin indices for mu on the slab
        mu = slab.mu(los) # defined with respect to specified LOS
        dig_mu = numpy.digitize(abs(mu).flat, muedges)

        # make the multi-index
        multi_index = numpy.ravel_multi_index([dig_x, dig_mu], (Nx+2,Nmu+2))

        # sum up x in each bin (accounting for negative freqs)
        xslab[:] *= slab.hermitian_weights
        xsum.flat += numpy.bincount(multi_index, weights=xslab.flat, minlength=xsum.size)

        # count number of modes in each bin (accounting for negative freqs)
        Nslab = numpy.ones_like(xslab) * slab.hermitian_weights
        Nsum.flat += numpy.bincount(multi_index, weights=Nslab.flat, minlength=Nsum.size)

        # compute multipoles by weighting by Legendre(ell, mu)
        for iell, ell in enumerate(_poles):

            # weight the input 3D array by the appropriate Legendre polynomial
            weighted_y3d = legpoly[iell](mu) * y3d[slab.index]

            # add conjugate for this kx, ky, kz, corresponding to
            # the (-kx, -ky, -kz) --> need to make mu negative for conjugate
            # Below is identical to the sum of
            # Leg(ell)(+mu) * y3d[:, nonsingular]    (kx, ky, kz)
            # Leg(ell)(-mu) * y3d[:, nonsingular].conj()  (-kx, -ky, -kz)
            # or
            # weighted_y3d[:, nonsingular] += (-1)**ell * weighted_y3d[:, nonsingular].conj()
            # but numerically more accurate.
            if hermitian_symmetric:

                if ell % 2: # odd, real part cancels
                    weighted_y3d.real[slab.nonsingular] = 0.
                    weighted_y3d.imag[slab.nonsingular] *= 2.
                else:  # even, imag part cancels
                    weighted_y3d.real[slab.nonsingular] *= 2.
                    weighted_y3d.imag[slab.nonsingular] = 0.

            # sum up the weighted y in each bin
            weighted_y3d *= (2.*ell + 1.)
            ysum[iell,...].real.flat += numpy.bincount(multi_index, weights=weighted_y3d.real.flat, minlength=Nsum.size)
            if numpy.iscomplexobj(ysum):
                ysum[iell,...].imag.flat += numpy.bincount(multi_index, weights=weighted_y3d.imag.flat, minlength=Nsum.size)

        # sum up the absolute mag of mu in each bin (accounting for negative freqs)
        mu[:] *= slab.hermitian_weights
        musum.flat += numpy.bincount(multi_index, weights=abs(mu).flat, minlength=musum.size)

    # sum binning arrays across all ranks
    xsum  = comm.allreduce(xsum)
    musum = comm.allreduce(musum)
    ysum  = comm.allreduce(ysum)
    Nsum  = comm.allreduce(Nsum)

    # add the last 'internal' mu bin (mu == 1) to the last visible mu bin
    # this makes the last visible mu bin inclusive on both ends.
    ysum[..., -2] += ysum[..., -1]
    musum[:, -2]  += musum[:, -1]
    xsum[:, -2]   += xsum[:, -1]
    Nsum[:, -2]   += Nsum[:, -1]

    # reshape and slice to remove out of bounds points
    sl = slice(1, -1)
    with numpy.errstate(invalid='ignore'):

        # 2D binned results
        y2d       = (ysum[0,...] / Nsum)[sl,sl] # ell=0 is first index
        xmean_2d  = (xsum / Nsum)[sl,sl]
        mumean_2d = (musum / Nsum)[sl, sl]
        N_2d      = Nsum[sl,sl]

        # 1D multipole results (summing over mu (last) axis)
        if do_poles:
            N_1d     = Nsum[sl,sl].sum(axis=-1)
            xmean_1d = xsum[sl,sl].sum(axis=-1) / N_1d
            poles    = ysum[:, sl,sl].sum(axis=-1) / N_1d
            poles    = poles[ell_idx,...]

    # return y(x,mu) + (possibly empty) multipoles
    result = (xmean_2d, mumean_2d, y2d, N_2d)
    pole_result = (xmean_1d, poles, N_1d) if do_poles else None
    return result, pole_result

def _cast_source(source, BoxSize, Nmesh):
    """
    Cast an object to a MeshSource. BoxSize and Nmesh is used
    only on CatalogSource
    """
    from pmesh.pm import Field
    from nbodykit.source.mesh import FieldMesh

    if isinstance(source, Field):
        # if input is a Field object, wrap it as a MeshSource.
        source = FieldMesh(source)
    elif isinstance(source, CatalogSourceBase):
        # if input is CatalogSource, use defaults to make it into a mesh
        if not isinstance(source, MeshSource):
            source = source.to_mesh(BoxSize=BoxSize, Nmesh=Nmesh, dtype='f8', compensated=True)

    # now we are having a MeshSource
    # paint the density field to the mesh
    if not isinstance(source, MeshSource):
        raise TypeError("Unknown type of source in FFTPower: %s" % str(type(source)))
    if BoxSize is not None and any(source.attrs['BoxSize'] != BoxSize):
        raise ValueError("Mismatched Boxsize between __init__ and source.attrs")
    if Nmesh is not None and any(source.attrs['Nmesh'] != Nmesh):
        raise ValueError(("Mismatched Nmesh between __init__ and source.attrs; "
                          "if trying to re-sample with a different mesh, specify "
                          "`Nmesh` as keyword of to_mesh()"))

    return source



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from numpy.testing import assert_allclose
import pytest

setup_logging()

@MPITest([4])
def test_missing_columns(comm):

    CurrentMPIComm.set(comm)

    # create FKP catalog
    source1 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=42)
    source2 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=84)
    cat = FKPCatalog(source1, source2, BoxSize=512.0, BoxPad=0.02)

    with pytest.raises(ValueError):
        mesh = cat.to_mesh(Nmesh=32)

@MPITest([4])
def test_boxsize(comm):

    CurrentMPIComm.set(comm)

    # data and randoms
    source1 = UniformCatalog(nbar=3e-3, BoxSize=512., seed=42)
    source2 = UniformCatalog(nbar=3e-3, BoxSize=512., seed=84)

    # add required columns
    source1['NZ']  = 1.0
    source2['NZ']  = 1.0

    # create FKPCatalog
    cat = FKPCatalog(source1, source2, BoxPad=0.02)

    # no Nmesh?
    with pytest.raises(ValueError):
        mesh = cat.to_mesh()

    # mesh
    mesh = cat.to_mesh(Nmesh=32)

    # check boxsize
    assert_allclose(mesh.attrs['BoxSize'], numpy.ceil(1.02*512.))



import numpy

def get_slice_size(start, stop, step):
    """
    Utility function to return the size of an array slice
    
    Parameters
    ----------
    start : int
        the beginning of the slice
    stop : int
        the end of the slice
    step : int
        the slice step size
    
    Returns
    -------
    N : int
        the total size of the slice
    """
    N, remainder = divmod(stop-start, step)
    if remainder: N += 1
    return N    

def global_to_local_slice(sizes, start, stop, fnum):
    """
    Convert a global slice, specified by `start` and `stop` 
    to the corresponding local indices of the file specified
    by `fnum`
    
    Parameters
    ----------
    sizes : array_like
        the sizes of each file in the file stack
    start : int
        the global index to begin the slice
    stop : int
        the global index to stop the slice
    fnum : int
        the file number that defines the desired local indexing
    
    Returns
    -------
    local_start, local_stop : int
        the local start and stop slice values
    """
    # cumulative sizes
    cumsizes = numpy.insert(numpy.cumsum(sizes), 0, 0)
    
    # local slice
    start_size = cumsizes[fnum] 
    return (max(start-start_size, 0), min(stop-start_size, sizes[fnum]))

def get_file_slice(sizes, start, stop):
    """
    Return the list of file numbers that must be accessed
    to return data between `start` and `slice`, where
    these indices are defined in terms of the global
    catalog indexing
    
    Parameters
    ----------
    sizes : array_like
        the sizes of each file in the file stack
    start : int
        the global index to begin the slice
    stop : int
        the global index to stop the slice
    
    Returns
    -------
    fnums : list
        the list of integers specifying the relevant
        file numbers that must be accessed
    """
    # cumulative sizes
    cumsizes = numpy.insert(numpy.cumsum(sizes), 0, 0)
    
    # return the relevant file numbers
    fnums = numpy.searchsorted(cumsizes[1:], [start, stop])
    return list(range(fnums[0], fnums[1]+1))


import os
import traceback
import logging
import numpy
from mpi4py import MPI
from nbodykit import CurrentMPIComm

def split_ranks(N_ranks, N, include_all=False):
    """
    Divide the ranks into chunks, attempting to have `N` ranks
    in each chunk. This removes the master (0) rank, such
    that `N_ranks - 1` ranks are available to be grouped

    Parameters
    ----------
    N_ranks : int
        the total number of ranks available
    N : int
        the desired number of ranks per worker
    include_all : bool, optional
        if `True`, then do not force each group to have
        exactly `N` ranks, instead including the remainder as well;
        default is `False`
    """
    available = list(range(1, N_ranks)) # available ranks to do work
    total = len(available)
    extra_ranks = total % N

    if include_all:
        for i, chunk in enumerate(numpy.array_split(available, max(total//N, 1))):
            yield i, list(chunk)
    else:
        for i in range(total//N):
            yield i, available[i*N:(i+1)*N]

        i = total // N
        if extra_ranks and extra_ranks >= N//2:
            remove = extra_ranks % 2 # make it an even number
            ranks = available[-extra_ranks:]
            if remove: ranks = ranks[:-remove]
            if len(ranks):
                yield i+1, ranks

def enum(*sequential, **named):
    """
    Enumeration values to serve as status tags passed
    between processeseee
    """
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type('Enum', (), enums)


class TaskManager(object):
    """
    An MPI task manager that distributes tasks over a set of MPI processes,
    using a specified number of independent workers to compute each task.

    Given the specified number of independent workers (which compute
    tasks in parallel), the total number of available CPUs will be
    divided evenly.

    The main function is ``iterate`` which iterates through a set of tasks,
    distributing the tasks in parallel over the available ranks.

    Parameters
    ----------
    cpus_per_task : int
        the desired number of ranks assigned to compute
        each task
    comm : MPI communicator, optional
        the global communicator that will be split so each worker
        has a subset of CPUs available; default is COMM_WORLD
    debug : bool, optional
        if `True`, set the logging level to `DEBUG`, which prints
        out much more information; default is `False`
    use_all_cpus : bool, optional
        if `True`, use all available CPUs, including the remainder
        if `cpus_per_task` is not divide the total number of CPUs
        evenly; default is `False`
    """
    logger = logging.getLogger('TaskManager')

    @CurrentMPIComm.enable
    def __init__(self, cpus_per_task, comm=None, debug=False, use_all_cpus=False):

        if debug:
            self.logger.setLevel(logging.DEBUG)

        self.cpus_per_task = cpus_per_task
        self.use_all_cpus  = use_all_cpus

        # the base communicator
        self.basecomm = MPI.COMM_WORLD if comm is None else comm
        self.rank     = self.basecomm.rank
        self.size     = self.basecomm.size

        # need at least one
        if self.size == 1:
            raise ValueError("need at least two processes to use a TaskManager")

        # communication tags
        self.tags = enum('READY', 'DONE', 'EXIT', 'START')

        # the task communicator
        self.comm = None

        # store a MPI status
        self.status = MPI.Status()

    def __enter__(self):
        """
        Split the base communicator such that each task gets allocated
        the specified number of cpus to perform the task with
        """
        chain_ranks = []
        color = 0
        total_ranks = 0
        nworkers = 0

        # split the ranks
        for i, ranks in split_ranks(self.size, self.cpus_per_task, include_all=self.use_all_cpus):
            chain_ranks.append(ranks[0])
            if self.rank in ranks: color = i+1
            total_ranks += len(ranks)
            nworkers = nworkers + 1
        self.workers = nworkers # store the total number of workers

        # check for no workers!
        if self.workers == 0:
            raise ValueError("no pool workers available; try setting `use_all_cpus` = True")

        leftover = (self.size - 1) - total_ranks
        if leftover and self.rank == 0:
            args = (self.cpus_per_task, self.size-1, leftover)
            self.logger.warning("with `cpus_per_task` = %d and %d available rank(s), %d rank(s) will do no work" %args)
            self.logger.warning("set `use_all_cpus=True` to use all available cpus")

        # crash if we only have one process or one worker
        if self.size <= self.workers:
            args = (self.size, self.workers+1, self.workers)
            raise ValueError("only have %d ranks; need at least %d to use the desired %d workers" %args)

        # ranks that will do work have a nonzero color now
        self._valid_worker = color > 0

        # split the comm between the workers
        self.comm = self.basecomm.Split(color, 0)
        self.original_comm = CurrentMPIComm.get()
        CurrentMPIComm.set(self.comm)

        return self

    def is_root(self):
        """
        Is the current process the root process?

        Root is responsible for distributing the tasks to the other available ranks
        """
        return self.rank == 0

    def is_worker(self):
        """
        Is the current process a valid worker?

        Workers wait for instructions from the master
        """
        try:
            return self._valid_worker
        except:
            raise ValeuError("workers are only defined when inside the ``with TaskManager()`` context")

    def _get_tasks(self):
        """
        Internal generator that yields the next available task from a worker
        """
        if self.is_root():
            raise RuntimeError("Root rank mistakenly told to await tasks")

        # logging info
        if self.comm.rank == 0:
            args = (self.rank, MPI.Get_processor_name(), self.comm.size)
            self.logger.debug("worker master rank is %d on %s with %d processes available" %args)

        # continously loop and wait for instructions
        while True:
            args = None
            tag = -1

            # have the master rank of the subcomm ask for task and then broadcast
            if self.comm.rank == 0:
                self.basecomm.send(None, dest=0, tag=self.tags.READY)
                args = self.basecomm.recv(source=0, tag=MPI.ANY_TAG, status=self.status)
                tag = self.status.Get_tag()

            # bcast to everyone in the worker subcomm
            args  = self.comm.bcast(args) # args is [task_number, task_value]
            tag   = self.comm.bcast(tag)

            # yield the task
            if tag == self.tags.START:

                # yield the task value
                yield args

                # wait for everyone in task group before telling master this task is done
                self.comm.Barrier()
                if self.comm.rank == 0:
                    self.basecomm.send([args[0], None], dest=0, tag=self.tags.DONE)

            # see ya later
            elif tag == self.tags.EXIT:
                break

        # wait for everyone in task group and exit
        self.comm.Barrier()
        if self.comm.rank == 0:
            self.basecomm.send(None, dest=0, tag=self.tags.EXIT)

        # debug logging
        self.logger.debug("rank %d process is done waiting" %self.rank)

    def _distribute_tasks(self, tasks):
        """
        Internal function that distributes the tasks from the root to the workers
        """
        if not self.is_root():
            raise ValueError("only the root rank should distribute the tasks")

        ntasks = len(tasks)
        task_index     = 0
        closed_workers = 0

        # logging info
        args = (self.workers, ntasks)
        self.logger.debug("master starting with %d worker(s) with %d total tasks" %args)

        # loop until all workers have finished with no more tasks
        while closed_workers < self.workers:

            # look for tags from the workers
            data = self.basecomm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=self.status)
            source = self.status.Get_source()
            tag = self.status.Get_tag()

            # worker is ready, so send it a task
            if tag == self.tags.READY:

                # still more tasks to compute
                if task_index < ntasks:
                    this_task = [task_index, tasks[task_index]]
                    self.basecomm.send(this_task, dest=source, tag=self.tags.START)
                    self.logger.debug("sending task `%s` to worker %d" %(str(tasks[task_index]), source))
                    task_index += 1

                # all tasks sent -- tell worker to exit
                else:
                    self.basecomm.send(None, dest=source, tag=self.tags.EXIT)

            # store the results from finished tasks
            elif tag == self.tags.DONE:
                self.logger.debug("received result from worker %d" %source)

            # track workers that exited
            elif tag == self.tags.EXIT:
                closed_workers += 1
                self.logger.debug("worker %d has exited, closed workers = %d" %(source, closed_workers))

    def iterate(self, tasks):
        """
        A generator that iterates through a series of tasks in parallel.

        Notes
        -----
        This is a collective operation and should be called by
        all ranks

        Parameters
        ----------
        tasks : iterable
            an iterable of `task` items that will be yielded in parallel
            across all ranks

        Yields
        -------
        task :
            the individual items of `tasks`, iterated through in parallel
        """
        # master distributes the tasks and tracks closed workers
        if self.is_root():
            self._distribute_tasks(tasks)

        # workers will wait for instructions
        elif self.is_worker():
            for tasknum, args in self._get_tasks():
                yield args

    def map(self, function, tasks):
        """
        Like the built-in :func:`map` function, apply a function to all
        of the values in a list and return the list of results.

        If ``tasks`` contains tuples, the arguments are passed to
        ``function`` using the ``*args`` syntax

        Notes
        -----
        This is a collective operation and should be called by
        all ranks

        Parameters
        ----------
        function : callable
            The function to apply to the list.
        tasks : list
            The list of tasks

        Returns
        -------
        results : list
            the list of the return values of :func:`function`
        """
        results = []

        # master distributes the tasks and tracks closed workers
        if self.is_root():
            self._distribute_tasks(tasks)

        # workers will wait for instructions
        elif self.is_worker():

            # iterate through tasks in parallel
            for tasknum, args in self._get_tasks():

                # make function arguments consistent with *args
                if not isinstance(args, tuple):
                    args = (args,)

                # compute the result (only worker root needs to save)
                result = function(*args)
                if self.comm.rank == 0:
                    results.append((tasknum, result))

        # put the results in the correct order
        results = self.basecomm.allgather(results)
        results = [item for sublist in results for item in sublist]
        return [r[1] for r in sorted(results, key=lambda x: x[0])]

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Exit gracefully by closing and freeing the MPI-related variables
        """
        if exc_value is not None:
            trace = ''.join(traceback.format_exception(exc_type, exc_value, exc_traceback, limit=5))
            self.logger.error("an exception has occurred on rank %d:\n%s" %(self.rank, trace))

            # bit of hack that forces mpi4py to exit all ranks
            # see https://groups.google.com/forum/embed/#!topic/mpi4py/RovYzJ8qkbc
            os._exit(1)

        # wait and exit
        self.logger.debug("rank %d process finished" %self.rank)
        self.basecomm.Barrier()

        if self.is_root():
            self.logger.debug("master is finished; terminating")

        CurrentMPIComm.set(self.original_comm)

        if self.comm is not None:
            self.comm.Free()



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_allclose
import tempfile
import os

@MPITest([1])
def test_hdf(comm):

    import h5py

    # fake structured array
    dset = numpy.empty(1024, dtype=[('Position', ('f8', 3)), ('Mass', 'f8')])
    dset['Position'] = numpy.random.random(size=(1024, 3))
    dset['Mass'] = numpy.random.random(size=1024)

    tmpfile = tempfile.mkstemp()[1]

    with h5py.File(tmpfile , 'w') as ff:
        ds = ff.create_dataset('X', data=dset) # store structured array as dataset
        ds.attrs['BoxSize'] = 1.0
        grp = ff.create_group('Y')
        grp.create_dataset('Position', data=dset['Position']) # column as dataset
        grp.create_dataset('Mass', data=dset['Mass']) # column as dataset

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    source = HDFCatalog(tmpfile, root='X', attrs={"Nmesh":32})
    assert_allclose(source['Position'], dset['Position'])

    os.unlink(tmpfile)

@MPITest([1])
def test_csv(comm):

    CurrentMPIComm.set(comm)

    with tempfile.NamedTemporaryFile() as ff:

        # generate data
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data, fmt='%.7e'); ff.seek(0)

        # read nrows
        names =['a', 'b', 'c', 'd', 'e']
        f = CSVCatalog(ff.name, names, blocksize=100)

        # make sure data is the same
        for i, name in enumerate(names):
            numpy.testing.assert_almost_equal(data[:,i], f[name].compute(), decimal=7)

        # make sure all the columns are there
        assert all(col in f for col in names)

@MPITest([1])
def test_stack_glob(comm):

    CurrentMPIComm.set(comm)
    tmpfile1 = 'test-glob-1.dat'
    tmpfile2 = 'test-glob-2.dat'

    # generate data
    data = numpy.random.random(size=(100,5))
    numpy.savetxt(tmpfile1, data, fmt='%.7e')
    numpy.savetxt(tmpfile2, data, fmt='%.7e')

    # read using a glob
    names =['a', 'b', 'c', 'd', 'e']
    f = CSVCatalog('test-glob-*', names, blocksize=100)

    # make sure print works
    print(f)

    # make sure data is the same
    fulldata = numpy.concatenate([data, data], axis=0)
    for i, name in enumerate(names):
        numpy.testing.assert_almost_equal(fulldata[:,i], f[name].compute(), decimal=7)

    # make sure all the columns are there
    assert all(col in f for col in names)

    os.unlink(tmpfile1)
    os.unlink(tmpfile2)

@MPITest([1])
def test_stack_list(comm):

    CurrentMPIComm.set(comm)
    tmpfile1 = 'test-list-1.dat'
    tmpfile2 = 'test-list-2.dat'

    # generate data
    data = numpy.random.random(size=(100,5))
    numpy.savetxt(tmpfile1, data, fmt='%.7e')
    numpy.savetxt(tmpfile2, data, fmt='%.7e')

    # read using a glob
    names =['a', 'b', 'c', 'd', 'e']
    f = CSVCatalog(['test-list-1.dat', 'test-list-2.dat'], names, blocksize=100)

    # make sure print works
    print(f)

    # make sure data is the same
    fulldata = numpy.concatenate([data, data], axis=0)
    for i, name in enumerate(names):
        numpy.testing.assert_almost_equal(fulldata[:,i], f[name].compute(), decimal=7)

    # make sure all the columns are there
    assert all(col in f for col in names)

    os.unlink(tmpfile1)
    os.unlink(tmpfile2)



import numpy

class MeshSlab(object):
    """
    A convenience class to represent a specific slab of a mesh,
    which is denoted as a ``slab``
    """
    def __init__(self, islab, coords, axis, symmetry_axis):
        """
        Parameters
        ----------
        islab : int, [0, x[0].shape[0]]
            the index of the slab, which indexes the first dimension of
            the mesh (the `x` coordinate), thus producing a y-z plane
        coords : list of arrays
            the coordinate arrays of the mesh, with proper 3D shapes
            for easy broadcasting; if the mesh has size (Nx, Ny, Nz),
            then the shapes of `x` are: [(Nx, 1, 1), (1, Ny, 1), (1, 1, Nz)]
        axis : int, {0, 1, 2}
            the index of the mesh axis to iterate over
        symmetry_axis : int, optional
            if provided, the axis that has been compressed due to Hermitian symmetry
        """
        self.ndim          = len(coords)
        self._index        = islab
        self.axis          = axis
        self.symmetry_axis = symmetry_axis
        self._coords       = coords

        # make sure symmetry_axis > 0 (sanity check)
        if self.hermitian_symmetric and self.symmetry_axis < 0:
            raise ValueError("`symmetry_axis` in MeshSlab must be non-negative")

    def __str__(self):
        name = self.__class__.__name__
        return "<%s: axis=%d, index=%d>" %(name, self.axis, self._index)

    def __repr__(self):
        return self.__str__()

    @property
    def index(self):
        """
        Return an indexing list appropriate for selecting the slicing the
        appropriate slab out of a full 3D array of shape (Nx, Ny, Nz)
        """
        toret = [slice(None)]*self.ndim
        toret[self.axis] = self._index
        return toret

    @property
    def meshshape(self):
        """
        Return the local shape of the mesh on this rank, as determined
        by the input coordinates array
        """
        return tuple([numpy.shape(self._coords[i])[i] for i in range(self.ndim)])

    @property
    def shape(self):
        """
        Return the shape of the slab
        """
        return tuple([s for i, s in enumerate(self.meshshape) if i != self.axis])

    @property
    def hermitian_symmetric(self):
        """
        Whether the slab is Hermitian-symmetric
        """
        return self.symmetry_axis is not None

    def coords(self, i):
        """
        Return the coordinate array for dimension ``i`` on this slab,

        .. note::

            The return value will be properly squeezed for easy
            broadcasting, i.e., if `i` is `self.axis`, then an array of
            shape `(1,1)` array is returned, otherwise, the shape is
            `(N_i, 1)` or `(1, N_i)`

        Parameters
        ----------
        i : int, {0,1,2}
            the index of the desired dimension

        Returns
        -------
        array_like
            the coordinate array for dimension `i` on the slab; see the
            note about the shape of the return array for details
        """
        if i < 0: i += self.ndim
        assert 0 <= i < self.ndim, "i should be between 0 and %d" %self.ndim

        if i != self.axis:
            return numpy.take(self._coords[i], 0, axis=self.axis)
        else:
            return numpy.take(self._coords[i], self._index, axis=self.axis)

    def norm2(self):
        """
        The square of coordinate grid norm defined at each
        point on the slab.

        This broadcasts the coordinate arrays along each dimension
        to compute the norm at each point in the slab.

        Returns
        -------
        array_like, (slab.shape)
            the square of coordinate mesh at each point in the slab
        """
        return sum(self.coords(i)**2 for i in range(self.ndim))

    def mu(self, los):
        """
        The `mu` value defined at each point on the slab for the
        specified line-of-sight index


        Parameters
        ----------
        los: array_like,
            the direction of the line-of-sight, which `mu` is defined
            with respect to; must have a norm of 1.

        Returns
        -------
        array_like, (slab.shape)
            the `mu` value at each point in the slab
        """
        with numpy.errstate(invalid='ignore'):
            return sum(self.coords(i) * los[i] for i in range(self.ndim)) / self.norm2()**0.5

    @property
    def nonsingular(self):
        """
        The indices on the slab of the positive frequencies
        along the dimension specified by `symmetry_axis`

        Returns
        -------
        idx : array_like, self.shape
            Return a boolean array with the shape of the slab,
            with `True` elements giving the elements with
            positive freq along `symmetry_axis`
        """
        try:
            return self._nonsingular
        except AttributeError:

            # initially, return slice that includes all elements
            idx = numpy.ones(self.shape, dtype=bool)

            # iteration axis is symmetry axis
            if self.symmetry_axis == self.axis:

                # check if current iteration value is positive
                if numpy.float(self.coords(self.axis)) <= 0.:
                    idx = numpy.zeros(self.shape, dtype=bool)

            # one of slab dimensions is symmetry axis
            else:

                # get the indices that have positive freq along symmetry axis
                nonsingular = (self._coords[self.symmetry_axis] > 0.)
                nonsingular = numpy.take(nonsingular, 0, axis=self.axis)

                idx[...] = nonsingular

            self._nonsingular = idx
            return self._nonsingular

    @property
    def hermitian_weights(self):
        """
        Weights to be applied to quantities on the slab in order
        to account for Hermitian symmetry

        These weights double-count the positive frequencies along
        the `symmetry_axis`.
        """
        try:
            return self._weights
        except AttributeError:

            # if not Hermitian symmetric, weights are 1
            if not self.hermitian_symmetric:
                toret = 1.
            # iteration axis is symmetry axis
            elif self.axis == self.symmetry_axis:
                toret = 1.
                if numpy.float(self.coords(self.symmetry_axis)) > 0.:
                    toret = 2.
            # only nonsingular plane gets factor of 2
            else:
                toret = numpy.ones(self.shape, dtype='f4')
                toret[self.nonsingular] = 2.

            self._weights = toret
            return self._weights

def SlabIterator(coords, axis=0, symmetry_axis=None):
    """
    Iterate over the specified dimension of the coordinate mesh,
    returning a :class:`MeshSlab` for each iteration

    Parameters
    ----------
    coords : list of arrays
        the coordinate arrays of the mesh, with proper 3D shapes
        for easy broadcasting; if the mesh has size (Nx, Ny, Nz),
        then the shapes of `x3d` should be: ``[(Nx, 1, 1), (1, Ny, 1), (1, 1, Nz)]``
    axis : int, optional
        the index of the mesh axis to iterate over
    symmetry_axis : int, optional
        if provided, the axis that has been compressed due to Hermitian symmetry
    """
    # number of dimensions in the mesh
    ndim = len(coords)
    if ndim not in [2,3]:
        raise NotImplementedError("SlabIterator can only be used on 3D or 2D arrays")

    # account for negative axes
    if axis < 0: axis += ndim
    assert 0 <= axis < ndim, "axis should be between 0 and %d" %ndim
    if symmetry_axis is not None and symmetry_axis < 0: symmetry_axis += ndim

    # this will only work if shapes are: [(Nx, 1, 1), (1, Ny, 1), (1, 1, Nz)]
    # mainly a sanity check to make sure things work
    shapes = [numpy.shape(x) for x in coords]
    try:
        mesh_size = [shape[i] for i, shape in enumerate(shapes)]
        for i in range(ndim):
            desired_shape = numpy.ones(ndim)
            desired_shape[i] = mesh_size[i]
            if shapes[i] != tuple(desired_shape):
                raise ValueError("coordinate array shape mismatch")
    except:
        raise ValueError("input coordinates with shapes %s are not correct" %str(shapes))

    # iterate over the specified axis, slab by slab
    N = numpy.shape(coords[axis])[axis]
    for islab in range(N):
        yield MeshSlab(islab, coords, axis, symmetry_axis)



import numpy
from .base import MPICorrfuncCallable, MissingCorrfuncError

class CorrfuncMocksCallable(MPICorrfuncCallable):
    """
    A MPI-enabled wrapper of a callable from :mod:`Corrfunc.mocks`.

    Parameters
    ----------
    func : callable
        the Corrfunc function that will be called
    edges : list
        the list of arrays specifying the bin edges in each coordinate direction
    """
    binning_dims = None

    def __init__(self, func, edges, show_progress=True):

        MPICorrfuncCallable.__init__(self, func, show_progress=show_progress)
        self.edges = edges

    def __call__(self, pos1, w1, pos2, w2, **config):

        kws = {}
        kws['autocorr'] = 0
        kws['nthreads'] = 1
        kws['binfile'] = self.edges[0]
        kws['RA2'] = pos2[:,0]
        kws['DEC2'] = pos2[:,1]
        kws['weights2'] = w2.astype(pos2.dtype)
        kws['weight_type'] = 'pair_product'
        kws['output_%savg' %self.binning_dims[0]] = True

        # add in the comoving distances if we have them
        threedims = pos1.shape[1] == 3
        if threedims:
            kws['CZ2'] = pos2[:,2]
            kws['cosmology'] = 1
            kws['is_comoving_dist'] = True

        # add in the additional config keywords
        kws.update(config)

        def callback(kws, chunk):
            kws['RA1'] = pos1[chunk][:,0]
            kws['DEC1'] = pos1[chunk][:,1]
            if threedims: kws['CZ1'] = pos1[chunk][:,2]
            kws['weights1'] = w1[chunk].astype(pos1.dtype)

        # compute the result
        sizes = self.comm.allgather(len(pos1))
        return MPICorrfuncCallable.__call__(self, sizes, kws, callback=callback)


class DDsmu_mocks(CorrfuncMocksCallable):
    """
    A MPI-enabled wrapper of :func:`Corrfunc.mocks.DDsmu_mocks.DDsmu_mocks`.
    """
    binning_dims = ['s', 'mu']

    def __init__(self, edges, Nmu, show_progress=True):
        try:
            from Corrfunc.mocks import DDsmu_mocks
        except ImportError:
            raise MissingCorrfuncError()

        self.Nmu = Nmu
        mu_edges = numpy.linspace(0., 1., Nmu+1)
        CorrfuncMocksCallable.__init__(self, DDsmu_mocks, [edges, mu_edges],
                                        show_progress=show_progress)

    def __call__(self, pos1, w1, pos2, w2, **config):
        config['nmu_bins'] = int(self.Nmu)
        config['mu_max'] = 1.0
        return CorrfuncMocksCallable.__call__(self, pos1, w1, pos2, w2, **config)

class DDtheta_mocks(CorrfuncMocksCallable):
    """
    A MPI-enabled wrapper of :func:`Corrfunc.mocks.DDtheta_mocks.DDtheta_mocks`.
    """
    binning_dims = ['theta']

    def __init__(self, edges, show_progress=True):
        try:
            from Corrfunc.mocks import DDtheta_mocks
        except ImportError:
            raise MissingCorrfuncError()

        CorrfuncMocksCallable.__init__(self, DDtheta_mocks, [edges],
                                        show_progress=show_progress)

class DDrppi_mocks(CorrfuncMocksCallable):
    """
    A MPI-enabled wrapper of :func:`Corrfunc.mocks.DDrppi_mocks.DDrppi_mocks`.
    """
    binning_dims = ['rp', 'pi']

    def __init__(self, edges, pimax, show_progress=True):
        try:
            from Corrfunc.mocks import DDrppi_mocks
        except ImportError:
            raise MissingCorrfuncError()

        self.pimax = pimax
        pi_bins = numpy.linspace(0, pimax, int(pimax)+1)
        CorrfuncMocksCallable.__init__(self, DDrppi_mocks, [edges, pi_bins],
                                        show_progress=show_progress)

    def __call__(self, pos1, w1, pos2, w2, **config):
        config['pimax'] = float(self.pimax)
        return CorrfuncMocksCallable.__call__(self, pos1, w1, pos2, w2, **config)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

import pytest

# debug logging
setup_logging("debug")


@MPITest([1])
def test_missing_ranks(comm):

    CurrentMPIComm.set(comm)
    cpus_per_task = 2
    with pytest.raises(ValueError):
        with TaskManager(cpus_per_task, debug=True, use_all_cpus=True) as tm:
            pass

@MPITest([2])
def test_no_workers(comm):

    CurrentMPIComm.set(comm)
    cpus_per_task = 2
    with pytest.raises(ValueError):
        with TaskManager(cpus_per_task, debug=True, use_all_cpus=False) as tm:
            pass


@MPITest([2, 4])
def test_iterate(comm):

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    cpus_per_task = 2
    with TaskManager(cpus_per_task, debug=True, use_all_cpus=True) as tm:

        try:
            for seed in tm.iterate([0, 1, 2]):

                # uniform particles
                source = UniformCatalog(nbar=3e-7, BoxSize=1380., seed=seed)

                # compute P(k,mu) and multipoles
                r = FFTPower(source, mode='2d', Nmesh=8, poles=[0,2,4])

                # and save
                output = "./test_batch_uniform_seed%d.json" % seed
                r.save(output)

        except Exception as e:
            print(e)
            raise

@MPITest([4])
def test_map(comm):

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    def fftpower(seed):

        # uniform particles
        source = UniformCatalog(nbar=3e-7, BoxSize=1380., seed=seed)

        # compute P(k,mu) and multipoles
        r = FFTPower(source, mode='2d', Nmesh=8, poles=[0,2,4])

        # and save
        output = "./test_batch_uniform_seed%d.json" % seed
        r.save(output)

        return seed

    cpus_per_task = 2
    with TaskManager(cpus_per_task, debug=True, use_all_cpus=False) as tm:

        try:
            seeds = [0, 1, 2]
            results = tm.map(fftpower, seeds)
            assert all(seeds[i] == results[i] for i in range(len(seeds)))
        except Exception as e:
            print(e)
            raise



import numpy
import logging
from nbodykit import CurrentMPIComm
from nbodykit.binned_statistic import BinnedStatistic

class MissingCorrfuncError(Exception):
    def __init__(self):
        msg = "use either ``conda install -c bccp corrfunc`` "
        msg += "or ``pip install git+git://github.com/nickhand/Corrfunc``"
        self.args = (msg,)

class CorrfuncResult(object):
    """
    A class used internally to hold the array-like result of a
    pair counting algorithm from :mod:`Corrfunc`.

    This class is useful for reducing pair count results in parallel
    while accounting for columns that are pair-weighted.

    Parameters
    ----------
    data : numpy.ndarray
        the numpy structured array result from :mod:`Corrfunc`
    """
    valid = ['weightavg', 'npairs', 'savg', 'ravg', 'thetaavg', 'rpavg']

    def __init__(self, data):

        # copy over the valid colums from the input result
        dtype = [(col, data.dtype[col]) for col in data.dtype.names if col in self.valid]
        self.data = numpy.zeros(data.shape, dtype=dtype)
        self.columns = self.data.dtype.names
        for col in self.columns:
            self.data[col] = data[col]

    def __getitem__(self, col):
        return self.data[col]

    @property
    def shape(self):
        return self.data.shape

    @property
    def dtype(self):
        return self.data.dtype

    def reshape(self, *args, **kwargs):
        self.data = self.data.reshape(*args, **kwargs)
        return self

    def __radd__(self, other):
        return self + other

    def __add__(self, other):

        # all columns are pair-weighted except for "npairs"
        data = numpy.empty_like(self.data)
        for col in self.columns:
            if col == 'npairs': continue
            data[col] = self[col]*self['npairs'] + other[col]*other['npairs']

        # just sum up "npairs"
        data['npairs'] = self['npairs'] + other['npairs']

        # normalize by total "npairs"
        idx = data['npairs'] > 0.
        for col in self.columns:
            if col == 'npairs': continue
            data[col][idx] /= data['npairs'][idx]

        return CorrfuncResult(data)

class MPICorrfuncCallable(object):
    """
    A base class to represent an MPI-enabled :mod:`Corrfunc` callable.

    This class adds the following functionality to the Corrfunc code:

    - If ``show_progress`` is ``True``, execute the function in chunks,
      logging to screen the progress along the way. This is useful for
      potentially long running pair counting jobs.
    - When calling the function, capture stdout/stderr and C-level output
      and if an error occurs, raise an exception with all generated output for
      the user.

    The relevant subclasses of this class are in :mod:`mocks` and :mod:`theory`.
    """
    binning_dims = None
    logger = logging.getLogger("MPICorrfuncCallable")

    @CurrentMPIComm.enable
    def __init__(self, callable, comm=None, show_progress=True):

        self.callable = callable
        self.comm = comm
        self.show_progress = show_progress

    def __call__(self, loads, kwargs, callback=None):
        """
        Calls :attr:`callable` in iterations, optionally
        calling ``callback`` before each iteration.

        This allows the :mod:`Corrfunc` function :attr:`callable` to be called
        in chunks, giving the user a progress report after each iteration.

        Parameters
        ----------
        loads : list of int
            the list of loads for every rank; this corresponds to the number
            of particles in the in ``A`` if we are correlating ``A`` x ``B``
        kwargs : dict
            the dictionary of arguments to pass to ``func``
        callback : callable, optional
            a callable takings ``kwargs`` as its first argument and a slice
            object as its second argument; this will be called first during
            each iteration

        Returns
        -------
        result : BinnedStatistic
            the total binned pair counting result
        """
        # the rank with the largest load
        largest_load = numpy.argmax(loads)

        # do the pair counting
        def run(chunk):
            if callback is not None:
                callback(kwargs, chunk)
            return self._run(self.callable, kwargs)

        # log the function start
        if self.comm.rank == 0:
            name = self.callable.__module__ + '.' + self.callable.__name__
            self.logger.info("calling function '%s'" % name)

        # number of iterations
        N = 10 if self.show_progress else 1

        # run in chunks
        pc = None
        chunks = numpy.array_split(range(loads[self.comm.rank]), N, axis=0)
        for i, chunk in enumerate(chunks):
            this_pc = run(chunk)
            if self.comm.rank == largest_load and self.show_progress:
                self.logger.info("%d%% done" % (N*(i+1)))

            # sum up the results
            pc = this_pc if pc is None else pc + this_pc

        # convert flattened 1D results to 2D array
        if len(self.edges) > 1:
            pc = pc.reshape((-1, len(self.edges[1])-1))

        # reduce the result across all ranks
        pc = self.comm.allreduce(pc)

        # the dimension names (use "r" instead of "s")
        dims = list(self.binning_dims) # make a copy here
        if 's' in dims: dims[dims.index('s')] = 'r'

        # make a new structured array
        dtype = numpy.dtype([(dims[0], 'f8'), ('npairs', 'u8'), ('weightavg', 'f8')])
        data = numpy.zeros(pc.shape, dtype=dtype)

        # copy over main results
        data[dims[0]] = pc[self.binning_dims[0]+'avg']
        data['npairs'] = pc['npairs']
        data['weightavg'] = pc['weightavg']

        # return the BinnedStatistic
        return BinnedStatistic(dims, self.edges, data, fields_to_sum=['npairs'])

    def _run(self, func, kws):
        """
        Internal function to run the wrapped :mod:`Corrfunc` function
        :attr:`func`, passing in the keywords specified by ``kws``.

        .. note::
            This hides all output from the Corrfunc function (stdout, stderr,
            and C-level output), unless an exception occurs.
        """
        from nbodykit.utils import captured_output

        try:
            # record progress capture output for everything but root
            with captured_output(self.comm, root=None) as (out, err):
                result = func(**kws)

        except Exception as e:
            # get the values Corrfunc logged to stdout and stderr
            stdout = out.getvalue(); stderr = err.getvalue()

            # log all of the output in a new exception
            name = func.__module__ + '.' + func.__name__
            msg = "calling the function '%s' failed, " % name
            msg += "likely due to issues with input data/parameters. "
            msg += "Open at issue at https://github.com/bccp/nbodykit/issues for further help.\n"
            msg += "exception: %s\n" % str(e)
            msg += "stdout: %s\n" % stdout
            msg += "stderr: %s" % stderr
            raise RuntimeError(msg)

        return CorrfuncResult(result)



from nbodykit.cosmology import Cosmology, LinearPower, HalofitPower, ZeldovichPower
from nbodykit.cosmology import EHPower, NoWiggleEHPower
import numpy
from numpy.testing import assert_allclose
import pytest

def test_bad_transfer():

    with pytest.raises(ValueError):
        Plin = LinearPower(Cosmology(), redshift=0., transfer="BAD")

def test_from_astropy():

    from astropy.cosmology import Planck15
    Plin = LinearPower(Planck15, redshift=0)
    Pk = Plin(0.1)

def test_deprecated_ehpower():

    c = Cosmology()
    with pytest.warns(FutureWarning):
        Plin1 = EHPower(c, redshift=0)
        Plin2 = LinearPower(c, 0., transfer='EisensteinHu')
        assert_allclose(Plin1(0.1), Plin2(0.1))

    with pytest.warns(FutureWarning):
        Plin1 = NoWiggleEHPower(c, redshift=0)
        Plin2 = LinearPower(c, 0., transfer='NoWiggleEisensteinHu')
        assert_allclose(Plin1(0.1), Plin2(0.1))

def test_large_scales():

    c = Cosmology()
    k = numpy.logspace(-5, -2, 100)

    # linear power
    Plin = LinearPower(c, redshift=0)

    # nonlinear power
    Pnl = HalofitPower(c, redshift=0)

    # zeldovich power
    Pzel = ZeldovichPower(c, redshift=0)

    assert_allclose(Plin(k), Pnl(k), rtol=1e-2)
    assert_allclose(Plin(k), Pzel(k), rtol=1e-2)

def test_linear_norm():

    # initialize the power
    c = Cosmology().match(sigma8=0.82)
    P = LinearPower(c, redshift=0, transfer='CLASS')

    # compute for array
    k = numpy.logspace(-3, numpy.log10(0.99*c.P_k_max), 100)
    Pk1 = P(k)

    # change sigma8
    P.sigma8 = 0.75
    Pk2 = P(k)
    assert_allclose(Pk1.max()/Pk2.max(), (0.82/0.75)**2, rtol=1e-2)

    # change redshift
    P.redshift = 0.55
    Pk3 = P(k)
    D2 = c.scale_independent_growth_factor(0.)
    D3 = c.scale_independent_growth_factor(0.55)
    assert_allclose(Pk2.max()/Pk3.max(), (D2/D3)**2, rtol=1e-2)

def test_linear():

    # initialize the power
    c = Cosmology().match(sigma8=0.82)
    P = LinearPower(c, redshift=0, transfer='CLASS')

    # check velocity dispersion
    assert_allclose(P.velocity_dispersion(), 5.898, rtol=1e-3)

    # test sigma8
    assert_allclose(P.sigma_r(8.), c.sigma8, rtol=1e-5)

    # change sigma8
    P.sigma8 = 0.80
    c = c.match(sigma8=0.80)

    assert_allclose(P.sigma_r(8.), P.sigma8, rtol=1e-5)

    # change redshift and test sigma8(z)
    P.redshift = 0.55
    assert_allclose(P.sigma_r(8.), c.sigma8_z(P.redshift), rtol=1e-5)

    # desired wavenumbers (in h/Mpc)
    k = numpy.logspace(-3, 2, 500)

    # initialize EH power
    P1 = LinearPower(c, redshift=0., transfer="CLASS")
    P2 = LinearPower(c, redshift=0., transfer='EisensteinHu')
    P3 = LinearPower(c, redshift=0., transfer='NoWiggleEisensteinHu')

    # check different transfers (very roughly)
    Pk1 = P1(k)
    Pk2 = P2(k)
    Pk3 = P3(k)
    assert_allclose(Pk1 / Pk1.max(), Pk2 / Pk2.max(), rtol=0.1)
    assert_allclose(Pk1 / Pk1.max(), Pk3 / Pk3.max(), rtol=0.1)

    # also try scalar
    Pk = P(0.1)

def test_halofit():

    # initialize the power
    c = Cosmology().match(sigma8=0.82)
    P = HalofitPower(c, redshift=0)

    # k is out of range
    with pytest.raises(ValueError):
        Pk = P(2*c.P_k_max)

    # compute for scalar
    Pk = P(0.1)

    # compute for array
    k = numpy.logspace(-3, numpy.log10(0.99*c.P_k_max), 100)
    Pk1 = P(k)

    # change sigma8
    P.sigma8 = 0.75
    Pk2 = P(k)
    assert_allclose(Pk1.max()/Pk2.max(), (0.82/0.75)**2, rtol=1e-2)

    # change redshift
    P.redshift = 0.55
    Pk3 = P(k)
    D2 = c.scale_independent_growth_factor(0.)
    D3 = c.scale_independent_growth_factor(0.55)
    assert_allclose(Pk2.max()/Pk3.max(), (D2/D3)**2, rtol=1e-2)

def test_zeldovich():

    # initialize the power
    c = Cosmology().match(sigma8=0.82)
    P = ZeldovichPower(c, redshift=0)

    # compute for scalar
    Pk = P(0.1)

    # compute for array
    k = numpy.logspace(-3, numpy.log10(0.99*c.P_k_max), 100)
    Pk1 = P(k)

    # change sigma8
    P.sigma8 = 0.75
    Pk2 = P(k)
    assert_allclose(Pk1.max()/Pk2.max(), (0.82/0.75)**2, rtol=1e-2)

    # change redshift
    P.redshift = 0.55
    Pk3 = P(k)
    D2 = c.scale_independent_growth_factor(0.)
    D3 = c.scale_independent_growth_factor(0.55)
    assert_allclose(Pk2.max()/Pk3.max(), (D2/D3)**2, rtol=1e-2)



from .base import PairCountBase, verify_input_sources
import numpy
import logging


class SurveyDataPairCount(PairCountBase):
    r"""
    Count (weighted) pairs of objects from a survey data catalog
    as a function of :math:`r`, :math:`(r,\mu)`, :math:`(r_p, \pi)`, or
    :math:`\theta` using the :mod:`Corrfunc` package.

    See the Notes below for the allowed coordinate dimensions.

    The default weighting scheme uses the product of the weights for each
    object in a pair.

    Results are computed when the class is inititalized. See the documenation
    of :func:`~SurveyDataPairCount.run` for the attributes storing the
    results.

    .. note::

        The algorithm expects the positions of particles from a survey catalog
        be the sky coordinates, right ascension and declination, and redshift.
        To compute pair counts in a simulation box using Cartesian
        coordinates, see :class:`~nbodykit.algorithms.SimulationBoxPairCount`.

    .. warning::
        The right ascension and declination columns should be specified
        in degrees.

    Parameters
    ----------
    mode : '1d', '2d', 'projected', 'angular'
        compute pair counts as a function of the specified coordinate basis;
        see the Notes section below for specifics
    first : CatalogSource
        the first source of particles, providing the 'Position' column
    edges : array_like
        the separation bin edges along the first coordinate dimension;
        depending on ``mode``, the options are :math:`r`, :math:`r_p`, or
        :math:`\theta`. Expected units for distances are :math:`\mathrm{Mpc}/h`
        and degrees for angles. Length of nbins+1
    cosmo : :class:`~nbodykit.cosmology.cosmology.Cosmology`, optional
        the cosmology instance used to convert redshift into comoving distance;
        this is required for all cases except ``mode='angular'``
    second : CatalogSource, optional
        the second source of particles to cross-correlate
    Nmu : int, optional
        the number of :math:`\mu` bins, ranging from 0 to 1; requred if
        ``mode='2d'``
    pimax : float, optional
        The maximum separation along the line-of-sight when ``mode='projected'``.
        Distances along the :math:`\pi` direction are binned with unit
        depth. For instance, if ``pimax=40``, then 40 bins will be created
        along the :math:`\pi` direction.
    ra : str, optional
        the name of the column in the source specifying the
        right ascension coordinates in units of degrees; default is 'RA'
    dec : str, optional
        the name of the column in the source specifying the declination
        coordinates; default is 'DEC'
    redshift : str, optional
        the name of the column in the source specifying the redshift
        coordinates; default is 'Redshift'
    weight : str, optional
        the name of the column in the source specifying the object weights
    show_progress : bool, optional
        if ``True``, perform the pair counting calculation in 10 iterations,
        logging the progress after each iteration; this is useful for
        understanding the scaling of the code
    domain_factor : int, optional
        the integer value by which to oversubscribe the domain decomposition
        mesh before balancing loads; this number can affect the distribution
        of loads on the ranks -- an optimal value will lead to balanced loads
    **config : key/value pairs
        additional keywords to pass to the :mod:`Corrfunc` function

    Notes
    -----
    This class can compute pair counts using several different coordinate
    choices, based on the value of the input argument ``mode``. The choices
    are:

    * ``mode='1d'`` : compute pairs as a function of the 3D separation :math:`r`
    * ``mode='2d'`` : compute pairs as a function of the 3D separation :math:`r`
      and the cosine of the angle to the line-of-sight, :math:`\mu`
    * ``mode='projected'`` : compute pairs as a function of distance perpendicular
      and parallel to the line-of-sight, :math:`r_p` and :math:`\pi`
    * ``mode='angular'`` : compute pairs as a function of angle on the sky, :math:`\theta`
    """
    logger = logging.getLogger('SurveyDataPairCount')

    def __init__(self, mode, first, edges, cosmo=None, second=None,
                    Nmu=None, pimax=None,
                    ra='RA', dec='DEC', redshift='Redshift', weight='Weight',
                    show_progress=False, domain_factor=4,
                    **config):

        # verify the input sources
        required_cols = [ra, dec, weight]
        if mode != 'angular': required_cols.append(redshift)
        verify_input_sources(first, second, None, required_cols, inspect_boxsize=False)

        # init the base class (this verifies input arguments)
        PairCountBase.__init__(self, mode, edges, first, second, Nmu, pimax, show_progress)

        # need cosmology if not angular!
        if mode != 'angular' and cosmo is None:
            raise ValueError("'cosmo' keyword is required when 'mode' is not 'angular'")

        # save the meta-data
        self.attrs['cosmo'] = cosmo
        self.attrs['weight'] = weight
        self.attrs['ra'] = ra
        self.attrs['dec'] = dec
        self.attrs['redshift'] = redshift
        self.attrs['config'] = config
        self.attrs['domain_factor'] = domain_factor

        # run the algorithm
        self.run()

    def run(self):
        """
        Calculate the pair counts of a survey data catalog.
        This adds the following attribute:

        - :attr:`SurveyDataPairCount.pairs`

        Attributes
        ----------
        pairs : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            a BinnedStatistic object holding the pair count results.
            The coordinate grid will be ``(r,)``, ``(r,mu)``, ``(rp, pi)``,
            or ``(theta,)`` when ``mode`` is '1d', '2d', 'projected', 'angular',
            respectively.

            The BinnedStatistic stores the following variables:

            - ``r``, ``rp``, or ``theta`` : the mean separation value in the bin
            - ``npairs``: the number of pairs in the bin
            - ``weightavg``: the average weight value in the bin; each pair
              contributes the product of the individual weight values
        """
        from .domain import decompose_survey_data

        # setup
        mode = self.attrs['mode']
        first, second = self.first, self.second
        attrs = self.attrs.copy()
        Nmu = 1 if mode == '1d' else attrs['Nmu']

        # compute the max cartesian distance for smoothing
        smoothing = numpy.max(attrs['edges'])
        if mode == 'projected':
            smoothing = numpy.sqrt(smoothing**2 + attrs['pimax']**2)
        elif mode == 'angular':
            smoothing = 2 * numpy.sin(0.5 * numpy.deg2rad(smoothing))

        # do a domain decomposition on the data
        (pos1, w1), (pos2, w2) = decompose_survey_data(first, second, attrs,
                                                        self.logger, smoothing,
                                                        angular=(mode=='angular'),
                                                        domain_factor=attrs['domain_factor'])

        # get the Corrfunc callable based on mode
        if attrs['mode'] in ['1d', '2d']:
            from .corrfunc.mocks import DDsmu_mocks
            func = DDsmu_mocks(attrs['edges'], Nmu, show_progress=attrs['show_progress'])

        elif attrs['mode'] == 'projected':
            from .corrfunc.mocks import DDrppi_mocks
            func = DDrppi_mocks(attrs['edges'], attrs['pimax'], show_progress=attrs['show_progress'])

        elif attrs['mode'] == 'angular':
            from .corrfunc.mocks import DDtheta_mocks
            func = DDtheta_mocks(attrs['edges'], show_progress=attrs['show_progress'])

        # do the calculation
        self.pairs = func(pos1, w1, pos2, w2, **attrs['config'])

        # squeeze the result if '1d' (single mu bin was used)
        if mode == '1d':
            self.pairs = self.pairs.squeeze()



import numpy as np
from scipy.integrate import odeint

class PerturbationGrowth(object):
    """
    Perturbation Growth coefficients at several orders.

    2-LPT is implemented. This implements the single fluid
    model of Boltamann equations. Therefore it is accurate
    only in a matter dominated universe. When background
    includes the radation contribution, the first order result
    is tuned to agree
    at sub-percent level comparing to a true multi-fluid
    boltzmann code under Planck15 cosmology.

    All derivatives are against ``lna``.

    .. note::
        Formulas are derived from Yin Li's notes on 2LPT.

    Parameters
    ----------
    cosmo: :class:`~nbodykit.cosmology.core.Cosmology`
        a astropy Cosmology like object.
    a : array_like
        a list of time steps where the factors are exact.
        other a values are interpolated.
    """
    def __init__(self, cosmo, a=None):
#        assert cosmo.Ogamma0 == 0
#        assert cosmo.Onu0 == 0

        self.cosmo = cosmo
        self.efunc = cosmo.efunc
        self.efunc_prime = cosmo.efunc_prime
        self.Om0 = cosmo.Om0
        self.Ogamma0 = cosmo.Ogamma0
        # suggested by Yin Li, not so sensitive to a_H
        # as long as it is early enough.
        # I used this number to ensure best agreement with CLASS
        # for Planck15. (1e-4 rtol)
        self.a_H = 5.22281250e-05

        if a is None:
            lna = np.log(np.logspace(-7, 0, 1024*10, endpoint=True))
        else:
            a = np.array(a, copy=True).ravel() # ensure this is 1-d
            if 1.0 not in a: # ensure redshift 0 is in the list, for normalization
                a = np.concatenate([[1.0], a])
            a.sort()
            if a[0] > 1e-7: # add a high redshift starting point.
                a = np.concatenate([[1e-7], a])
            lna = np.log(a)

        self.lna = lna
        self._D1, self._D2 = self._solve()

    def D1(self, a, order=0):
        """ Linear order growth function.

            Parameters
            ----------
            a : float, array_like
                scaling factor
            order : int
                order of differentation; 1 for first derivative against log a.

            Returns
            -------
            array_like : linear order growth function.
        """
        lna = np.log(a)
        return np.interp(lna, self.lna, self._D1[:, order])

    def D2(self, a, order=0):
        """ Second order growth function.

            Parameters
            ----------
            a : float, array_like
                scaling factor
            order : int
                order of differentation; 1 for first derivative against log a.

            Returns
            -------
            array_like : second order growth function.
        """
        lna = np.log(a)
        return np.interp(lna, self.lna, self._D2[:, order])

    def f1(self, a):
        """ Linear order growth rate

            Parameters
            ----------
            a : float, array_like
                scaling factor
            order : int
                order of differentation; 1 for first derivative against log a.

            Returns
            -------
            array_like : linear order growth rate.
        """
        return self.D1(a, order=1) / self.D1(a, order=0)

    def f2(self, a):
        """ Second order growth rate.

            Parameters
            ----------
            a : float, array_like
                scaling factor
            order : int
                order of differentation; 1 for first derivative against log a.

            Returns
            -------
            array_like : second order growth rate.
        """
        return self.D2(a, order=1) / self.D2(a, order=0)

    def Gp(self, a):
        """
        FastPM growth factor function, eq, 19
        """
        return self.D1(a)

    def gp(self, a):
        """
        Notice the derivative of D1 is against ln a but gp is d D1 / da, so
        gp = D1(a, order=1) / a
        """
        return self.D1(a, order=1) / a

    def Gf(self, a):
        """
        FastPM growth factor function, eq, 20
        """
        return self.D1(a, 1) * a ** 2 * self.E(a)

    def gf(self, a):
        """
        Similarly, the derivative is against ln a, so
        gf = Gf(a, order=1) / a
        """
        return 1 / a * (
            self.D1(a, 2) * a ** 2 * self.E(a) \
            +  self.D1(a, 1) * (
                    a ** 2 * self.E(a, order=1)
                +   2 * a ** 2 * self.E(a))
            )

    def E(self, a, order=0):
        """ Hubble function and derivatives against log a.

        """
        if order == 0:
            return self.efunc(1./a - 1.0)
        else:
            return self.efunc_prime(1./a - 1.0) * a

    def Hfac(self, a):
        return -2. - self.E(a, order=1) / self.E(a)

    def Om(self, a):
        z = 1./a-1
        return self.cosmo.Omega_b(z) + self.cosmo.Omega_cdm(z) # non-relativistic 

    def ode(self, y, lna):
        D1, F1, D2, F2 = y
        a = np.exp(lna)
        hfac = self.Hfac(a)
        omega = self.Om(a)
        F1p = hfac * F1 + 1.5 * omega * D1
        D1p = F1
        F2p = hfac * F2 + 1.5 * omega * D2 - 1.5 * omega * D1 ** 2
        D2p = F2
        return D1p, F1p, D2p, F2p

    def _solve(self):
        a0 = np.exp(self.lna[0])
        Om = self.Om(a0)
        if Om > 0.99:
            # matter dominated initial conditions
            y0 = [a0, a0, -3./7 * a0**2, -6. / 7 *a0**2]
        elif Om < 0.01:
            a_H = self.a_H
            D1i = np.log(a0 / a_H)
            # radiation dominated initial conditions
            y0 = [
                D1i,
                1.0,
                 -1.5 * Om * (D1i ** 2 - 4 * D1i + 6),
                 -1.5 * Om * (D1i ** 2 - 2 * D1i + 2)]
        else:
            raise ValueError('Neither matter or radiation dominated initial condition. Om(a_i) = %g.' % Om
                           + 'This shall not happen for a reasonable cosmology.')

        # solve with critical point at a=1.0, lna=0.
        y = odeint(self.ode, y0, self.lna, tcrit=[0.], atol=0)

        v1 = []
        v2 = []
        for yi, lnai in zip(y, self.lna):
            D1, F1, D2, F2 = yi
            D1p, F1p, D2p, F2p = self.ode(yi, lnai)
            v1.append((D1, F1, F1p))
            v2.append((D2, F2, F2p))

        v1 = np.array(v1)
        v2 = np.array(v2)

        ind = abs(self.lna).argmin()
        # normalization to 1 at a=1.0
        v1 /= v1[ind][0]
        v2 /= v2[ind][0]
        return v1, v2



import numpy
from .base import MPICorrfuncCallable, MissingCorrfuncError

class CorrfuncTheoryCallable(MPICorrfuncCallable):
    """
    A MPI-enabled wrapper of a callable from :mod:`Corrfunc.theory`.

    Parameters
    ----------
    func : callable
        the Corrfunc function that will be called
    edges : list
        the list of arrays specifying the bin edges in each coordinate direction
    periodic : bool
        whether to use periodic boundary conditions
    BoxSize : array_like
        the size of the simulation box in each direction
    """
    binning_dims = None

    def __init__(self, func, edges, periodic, BoxSize, show_progress=True):
        MPICorrfuncCallable.__init__(self, func, show_progress=show_progress)
        self.edges = edges
        self.periodic = periodic

        # check periodic/BoxSize compatibiltity
        if periodic:
            if not numpy.all(BoxSize == BoxSize[0]):
                raise NotImplementedError("periodic wrapping with non-cubic boxes not implemented yet")
            else:
                self.BoxSize = BoxSize[0] # isotropic box
        else:
            self.BoxSize = None


    def __call__(self, pos1, w1, pos2, w2, **config):

        kws = {}
        kws['autocorr'] = 0
        kws['nthreads'] = 1
        kws['binfile'] = self.edges[0]
        kws['X2'] = pos2[:,0]
        kws['Y2'] = pos2[:,1]
        kws['Z2'] = pos2[:,2] # the LOS direction
        kws['weights2'] = w2.astype(pos2.dtype)
        kws['weight_type'] = 'pair_product'
        kws['output_%savg' %self.binning_dims[0]] = True
        kws['periodic'] = self.periodic
        if self.BoxSize is not None: kws['boxsize'] = self.BoxSize

        # add in the additional config keywords
        kws.update(config)

        def callback(kws, chunk):
            kws['X1'] = pos1[chunk][:,0]
            kws['Y1'] = pos1[chunk][:,1]
            kws['Z1'] = pos1[chunk][:,2] # LOS defined with respect to this axis
            kws['weights1'] = w1[chunk].astype(pos1.dtype)

        # compute the result
        sizes = self.comm.allgather(len(pos1))
        return MPICorrfuncCallable.__call__(self, sizes, kws, callback=callback)

class DD(CorrfuncTheoryCallable):
    """
    A MPI-enabled wrapper of :func:`Corrfunc.theory.DD.DD`.
    """
    binning_dims = ['r']

    def __init__(self, edges, periodic, BoxSize, show_progress=True):
        try:
            from Corrfunc.theory import DD
        except ImportError:
            raise MissingCorrfuncError()

        CorrfuncTheoryCallable.__init__(self, DD, [edges], periodic, BoxSize,
                                        show_progress=show_progress)

class DDsmu(CorrfuncTheoryCallable):
    """
    A MPI-enabled wrapper of :func:`Corrfunc.theory.DDsmu.DDsmu`.
    """
    binning_dims = ['s', 'mu']

    def __init__(self, edges, Nmu, periodic, BoxSize, show_progress=True):
        try:
            from Corrfunc.theory import DDsmu
        except ImportError:
            raise MissingCorrfuncError()

        self.Nmu = Nmu
        mu_edges = numpy.linspace(0., 1., Nmu+1)
        CorrfuncTheoryCallable.__init__(self, DDsmu, [edges, mu_edges],
                                        periodic, BoxSize, show_progress=show_progress)

    def __call__(self, pos1, w1, pos2, w2, **config):
        config['nmu_bins'] = self.Nmu
        config['mu_max'] = 1.0
        return CorrfuncTheoryCallable.__call__(self, pos1, w1, pos2, w2, **config)

class DDrppi(CorrfuncTheoryCallable):
    """
    A MPI-enabled wrapper of :func:`Corrfunc.theory.DDrppi.DDrppi`.
    """
    binning_dims = ['rp', 'pi']

    def __init__(self, edges, pimax, periodic, BoxSize, show_progress=True):
        try:
            from Corrfunc.theory import DDrppi
        except ImportError:
            raise MissingCorrfuncError()

        self.pimax = pimax
        pi_bins = numpy.linspace(0, pimax, pimax+1)
        CorrfuncTheoryCallable.__init__(self, DDrppi, [edges, pi_bins],
                                        periodic, BoxSize, show_progress=show_progress)

    def __call__(self, pos1, w1, pos2, w2, **config):
        config['pimax'] = self.pimax
        return CorrfuncTheoryCallable.__call__(self, pos1, w1, pos2, w2, **config)



from nbodykit.base.catalog import CatalogSource, column
from nbodykit import cosmology
from nbodykit.utils import attrs_to_dict
from nbodykit import CurrentMPIComm

import numpy

class LogNormalCatalog(CatalogSource):
    """
    A CatalogSource containing biased particles that have
    been Poisson-sampled from a log-normal density field.

    Parameters
    ----------
    Plin : callable
        callable specifying the linear power spectrum
    nbar : float
        the number density of the particles in the box, assumed constant across
        the box; this is used when Poisson sampling the density field
    BoxSize : float, 3-vector of floats
        the size of the box to generate the grid on
    Nmesh : int
        the mesh size to use when generating the density and displacement
        fields, which are Poisson-sampled to particles
    bias : float, optional
        the desired bias of the particles; applied while applying a log-normal
        transformation to the density field
    seed : int, optional
        the global random seed; if set to ``None``, the seed will be set
        randomly
    cosmo : :class:`nbodykit.cosmology.core.Cosmology`, optional
        this must be supplied if ``Plin`` does not carry ``cosmo`` attribute
    redshift : float, optional
        this must be supplied if ``Plin`` does not carry a ``redshift`` attribute
    comm : MPI Communicator, optional
        the MPI communicator instance; default (``None``) sets to the
        current communicator

    References
    ----------
    `Cole and Jones, 1991 <http://adsabs.harvard.edu/abs/1991MNRAS.248....1C>`_
    `Agrawal et al. 2017 <https://arxiv.org/abs/1706.09195>`_
    """
    def __repr__(self):
        return "LogNormalCatalog(seed=%(seed)d, bias=%(bias)g)" %self.attrs

    @CurrentMPIComm.enable
    def __init__(self, Plin, nbar, BoxSize, Nmesh, bias=2., seed=None,
                    cosmo=None, redshift=None,
                    unitary_amplitude=False, inverted_phase=False, comm=None):

        self.comm = comm
        self.Plin = Plin

        # try to infer cosmo or redshift from Plin
        if cosmo is None:
            cosmo = getattr(self.Plin, 'cosmo', None)
        if redshift is None:
            redshift = getattr(self.Plin, 'redshift', None)
        if cosmo is None:
            raise ValueError("'cosmo' must be passed if 'Plin' does not have 'cosmo' attribute")
        if redshift is None:
            raise ValueError("'redshift' must be passed if 'Plin' does not have 'redshift' attribute")
        self.cosmo = cosmo

        # try to add attrs from the Plin
        if hasattr(Plin, 'attrs'):
            self.attrs.update(Plin.attrs)
        else:
            self.attrs['cosmo'] = dict(cosmo)

        # save the meta-data
        self.attrs['nbar']  = nbar
        self.attrs['redshift'] = redshift
        self.attrs['bias'] = bias
        self.attrs['unitary_amplitude'] = unitary_amplitude
        self.attrs['inverted_phase'] = inverted_phase

        # set the seed randomly if it is None
        if seed is None:
            if self.comm.rank == 0:
                seed = numpy.random.randint(0, 4294967295)
            seed = self.comm.bcast(seed)
        self.attrs['seed'] = seed

        # make the actual source
        self._source, pm = self._makesource(BoxSize=BoxSize, Nmesh=Nmesh)
        self.attrs['Nmesh'] = pm.Nmesh.copy()
        self.attrs['BoxSize'] = pm.BoxSize.copy()

        # set the size
        self._size = len(self._source)

        # init the base class
        CatalogSource.__init__(self, comm=comm)

        # crash with no particles!
        if self.csize == 0:
            raise ValueError("no particles in LogNormal source; try increasing ``nbar`` parameter")

    @column
    def Position(self):
        """
        Position assumed to be in Mpc/h
        """
        return self.make_column(self._source['Position'])

    @column
    def Velocity(self):
        """
        Velocity in km/s
        """
        return self.make_column(self._source['Velocity'])

    @column
    def VelocityOffset(self):
        """
        The corresponding RSD offset, in Mpc/h
        """
        return self.make_column(self._source['VelocityOffset'])

    def _makesource(self, BoxSize, Nmesh):

        from nbodykit import mockmaker
        from pmesh.pm import ParticleMesh

        # the particle mesh for gridding purposes
        _Nmesh = numpy.empty(3, dtype='i8')
        _Nmesh[:] = Nmesh
        pm = ParticleMesh(BoxSize=BoxSize, Nmesh=_Nmesh, dtype='f4', comm=self.comm)

        # growth rate to do RSD in the Zel'dovich approx
        f = self.cosmo.scale_independent_growth_rate(self.attrs['redshift'])

        # compute the linear overdensity and displacement fields
        delta, disp = mockmaker.gaussian_real_fields(pm, self.Plin, self.attrs['seed'],
                    unitary_amplitude=self.attrs['unitary_amplitude'],
                    inverted_phase=self.attrs['inverted_phase'],
                    compute_displacement=True)

        # poisson sample to points
        # this returns position and velocity offsets
        kws = {'bias':self.attrs['bias'], 'seed':self.attrs['seed'], 'comm':self.comm}
        pos, disp = mockmaker.poisson_sample_to_points(delta, disp, pm, self.attrs['nbar'], **kws)

        # move particles from initial position based on the Zeldovich displacement
        pos[:] = (pos + disp) % BoxSize

        # velocity from displacement (assuming Mpc/h)
        # this is f * H(z) * a / h = f 100 E(z) a --> converts from Mpc/h to km/s
        z = self.attrs['redshift']
        velocity_norm = f * 100 * self.cosmo.efunc(z) / (1+z)
        vel = velocity_norm * disp

        # return data
        dtype = numpy.dtype([
                ('Position', ('f4', 3)),
                ('Velocity', ('f4', 3)),
                ('VelocityOffset', ('f4', 3))
        ])
        source = numpy.empty(len(pos), dtype)
        source['Position'][:] = pos[:] # in Mpc/h
        source['Velocity'][:] = vel[:] # in km/s
        source['VelocityOffset'][:] = f*disp[:] # in Mpc/h

        return source, pm



from runtests.mpi import MPITest
from nbodykit import setup_logging
from nbodykit.binned_statistic import BinnedStatistic

import pytest
import tempfile
import numpy.testing as testing
import numpy
import os

data_dir = os.path.join(os.path.split(os.path.abspath(__file__))[0], 'data')
setup_logging("debug")


@MPITest([1])
def test_to_json(comm):

    # load from JSON
    ds1 = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_1d.json'))

    # to JSON
    with tempfile.NamedTemporaryFile(delete=False) as ff:
        ds1.to_json(ff.name)
    ds2 = BinnedStatistic.from_json(ff.name)

    # same data?
    for name in ds1:
        testing.assert_almost_equal(ds1[name], ds2[name])

    # cleanup
    os.remove(ff.name)

@MPITest([1])
def test_1d_load(comm):

    # load plaintext format
    with pytest.warns(FutureWarning):
        ds1 = BinnedStatistic.from_plaintext(['k'], os.path.join(data_dir, 'dataset_1d_deprecated.dat'))

        # wrong dimensions
        with pytest.raises(ValueError):
            ds1 = BinnedStatistic.from_plaintext(['k', 'mu'], os.path.join(data_dir, 'dataset_1d_deprecated.dat'))

    # load from JSON
    ds2 = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_1d.json'))

    # same data?
    for name in ds1:
        testing.assert_almost_equal(ds1[name], ds2[name])

@MPITest([1])
def test_2d_load(comm):

    # load plaintext format
    with pytest.warns(FutureWarning):
        ds1 = BinnedStatistic.from_plaintext(['k', 'mu'], os.path.join(data_dir, 'dataset_2d_deprecated.dat'))

    # load from JSON
    ds2 = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_2d.json'))

    # same data?
    for name in ds1:
        testing.assert_almost_equal(ds1[name], ds2[name])

@MPITest([1])
def test_str(comm):

    dataset = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_2d.json'))

    # list all variable names
    s = str(dataset)

    # now just list total number of variables
    dataset['test1'] = numpy.ones(dataset.shape)
    dataset['test2'] = numpy.ones(dataset.shape)
    s = str(dataset)

    # this is the same as str
    r = repr(dataset)

@MPITest([1])
def test_getitem(comm):

    dataset = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_2d.json'))

    # invalid key
    with pytest.raises(KeyError):
        bad = dataset['error']

    # slice columns
    sliced = dataset[['k', 'mu', 'power']]
    sliced = dataset[('k', 'mu', 'power')]

    # invalid slice
    with pytest.raises(KeyError):
        bad =dataset[['k', 'mu', 'error']]

    # too many dims in slice
    with pytest.raises(IndexError):
        bad = dataset[0,0,0]

    # cannot access single element of 2D power
    with pytest.raises(IndexError):
        bad = dataset[0,0]

@MPITest([1])
def test_array_slice(comm):

    dataset = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_2d.json'))

    # get the first mu column
    sliced = dataset[:,0]
    assert sliced.shape[0] == dataset.shape[0]
    assert len(sliced.shape) == 1
    assert sliced.dims == ['k']

    # get the first mu column but keep dimension
    sliced = dataset[:,[0]]
    assert sliced.shape[0] == dataset.shape[0]
    assert sliced.shape[1] == 1
    assert sliced.dims == ['k', 'mu']


@MPITest([1])
def test_list_array_slice(comm):

    dataset = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_2d.json'))

    # get the first and last mu column
    sliced = dataset[:,[0, -1]]
    assert len(sliced.shape) == 2
    assert sliced.dims == ['k', 'mu']

    # make sure we grabbed the right data
    for var in dataset:
        testing.assert_array_equal(dataset[var][:,[0,-1]], sliced[var])


@MPITest([1])
def test_variable_set(comm):

    dataset = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_2d.json'))
    modes = numpy.ones(dataset.shape)

    # add new variable
    dataset['TEST'] = modes
    assert 'TEST' in dataset

    # override existing variable
    dataset['modes'] = modes
    assert numpy.all(dataset['modes'] == 1.0)

    # needs right shape
    with pytest.raises(ValueError):
        dataset['TEST'] = 10.


@MPITest([1])
def test_copy(comm):

    dataset = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_2d.json'))
    copy = dataset.copy()
    for var in dataset:
        testing.assert_array_equal(dataset[var], copy[var])

@MPITest([1])
def test_rename_variable(comm):

    dataset = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_2d.json'))
    test = numpy.zeros(dataset.shape)
    dataset['test'] = test

    dataset.rename_variable('test', 'renamed_test')
    assert 'renamed_test' in dataset
    assert 'test' not in dataset

@MPITest([1])
def test_sel(comm):

    dataset = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_2d.json'))

    # no exact match fails
    with pytest.raises(IndexError):
        sliced = dataset.sel(k=0.1)

    # this should be squeezed
    sliced = dataset.sel(k=0.1, method='nearest')
    assert len(sliced.dims) == 1

    # this is not squeezed
    sliced = dataset.sel(k=[0.1], method='nearest')
    assert sliced.shape[0] == 1

    # slice in a specific k-range
    sliced = dataset.sel(k=slice(0.02, 0.15), mu=[0.5], method='nearest')
    assert sliced.shape[1] == 1
    assert numpy.alltrue((sliced['k'] >= 0.02)&(sliced['k'] <= 0.15))


@MPITest([1])
def test_squeeze(comm):

    dataset = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_2d.json'))

    # need to specify which dimension to squeeze
    with pytest.raises(ValueError):
        squeezed = dataset.squeeze()
    with pytest.raises(ValueError):
        squeezed = dataset[[0],[0]].squeeze()

    sliced = dataset[:,[2]]
    with pytest.raises(ValueError):
        squeezed = sliced.squeeze('k')
    squeezed = sliced.squeeze('mu')

    assert len(squeezed.dims) == 1
    assert squeezed.shape[0] == sliced.shape[0]


@MPITest([1])
def test_average(comm):
    import warnings

    dataset = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_2d.json'))

    # unweighted
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)

        avg = dataset.average('mu')
        for var in dataset.variables:
            if var in dataset._fields_to_sum:
                x = numpy.nansum(dataset[var], axis=-1)
            else:
                x = numpy.nanmean(dataset[var], axis=-1)
            testing.assert_allclose(x, avg[var])

        # weighted
        weights = numpy.random.random(dataset.shape)
        dataset['weights'] = weights
        avg = dataset.average('mu', weights='weights')

        for var in dataset:
            if var in dataset._fields_to_sum:
                x = numpy.nansum(dataset[var], axis=-1)
            else:
                x = numpy.nansum(dataset[var]*dataset['weights'], axis=-1)
                x /= dataset['weights'].sum(axis=-1)
            testing.assert_allclose(x, avg[var])


@MPITest([1])
def test_reindex(comm):
    import warnings

    dataset = BinnedStatistic.from_json(os.path.join(data_dir, 'dataset_2d.json'))

    with pytest.raises(ValueError):
        new, spacing = dataset.reindex('k', 0.005, force=True, return_spacing=True)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)

        weights = numpy.random.random(dataset.shape)
        dataset['weights'] = weights
        new, spacing = dataset.reindex('k', 0.02, weights='weights', force=True, return_spacing=True)

        diff = numpy.diff(new.coords['k'])
        assert numpy.alltrue(diff > numpy.diff(dataset.coords['k'])[0])

        with pytest.raises(ValueError):
            new = dataset.reindex('mu', 0.4, force=False)
        new = dataset.reindex('mu', 0.4, force=True)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_allclose, assert_array_equal

# debug logging
setup_logging("debug")

@MPITest([1])
def test_save(comm):

    N = 1000
    FSKY = 1.0
    
    CurrentMPIComm.set(comm)
    cosmo = cosmology.Planck15
    
    # create the source
    source = RandomCatalog(N, seed=42)
    source['z'] = source.rng.normal(loc=0.5, scale=0.1, size=source.size)
    
    # compute the histogram
    r = RedshiftHistogram(source, FSKY, cosmo, redshift='z')
    r.run()    
    r.save('zhist-test.json')

    r2 = RedshiftHistogram.load('zhist-test.json')

    assert_array_equal(r.bin_edges, r.bin_edges)
    assert_array_equal(r.bin_centers, r2.bin_centers)
    assert_array_equal(r.dV, r2.dV)
    assert_array_equal(r.nbar, r2.nbar)
    for k in r.attrs:
        assert_array_equal(r.attrs[k], r2.attrs[k])

@MPITest([1, 4])
def test_unweighted(comm):
    
    N = 1000
    FSKY = 1.0
    
    CurrentMPIComm.set(comm)
    cosmo = cosmology.Planck15
    
    # create the source
    source = RandomCatalog(N, seed=42)
    source['z'] = source.rng.normal(loc=0.5, scale=0.1, size=source.size)
    
    # compute the histogram
    r = RedshiftHistogram(source, FSKY, cosmo, redshift='z')

    assert (r.nbar*r.dV).sum() == N
    

@MPITest([1, 4])
def test_weighted(comm):
    
    N = 1000
    FSKY = 1.0
    
    CurrentMPIComm.set(comm)
    cosmo = cosmology.Planck15
    
    # create the source
    source = RandomCatalog(N, seed=42)
    source['z'] = source.rng.normal(loc=0.5, scale=0.1, size=source.size)
    source['weight'] = source.rng.uniform(0, high=1., size=source.size)
    
    # compute the histogram
    r = RedshiftHistogram(source, FSKY, cosmo, redshift='z', weight='weight')

    wsum = comm.allreduce(source['weight'].sum().compute())
    assert_allclose((r.nbar*r.dV).sum(), wsum)




    





from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import set_options
from nbodykit import setup_logging
from numpy.testing import assert_array_equal, assert_allclose
import pytest

# debug logging
setup_logging("debug")


@MPITest([1, 4])
def test_gslice(comm):
    CurrentMPIComm.set(comm)

    # the catalog to slice
    d = UniformCatalog(1000, 1.0, seed=42)
    d = d.to_mesh(Nmesh=32)

    sels = [(0,10,1), (None,10,1), (0,50,4), (0,50,-1), (-10,-1,1), (-10,None,1)]
    for (start,stop,end) in sels:

        sliced = d.gslice(start,stop,end)

        for col in d:
            data1 = numpy.concatenate(comm.allgather(d[col]), axis=0)
            data2 = numpy.concatenate(comm.allgather(sliced[col]), axis=0)

            sl = slice(start,stop,end)
            assert_array_equal(data1[sl], data2, err_msg="using slice= "+str(sl))

    # empty slice
    sliced = d.gslice(0,0)
    assert len(sliced) == 0

@MPITest([1, 4])
def test_sort_ascending(comm):
    CurrentMPIComm.set(comm)

    # the mesh to sort
    d = UniformCatalog(100, 1.0)
    d['mass'] = 10**(d.rng.uniform(low=12, high=15, size=d.size))
    mesh = d.to_mesh(Nmesh=32)

    # invalid sort key
    with pytest.raises(ValueError):
        mesh2 = mesh.sort('BadColumn', reverse=False)

    # duplicate sort keys
    with pytest.raises(ValueError):
        mesh2 = mesh.sort(['mass', 'mass'])

    # sort in ascending order by mass
    mesh2 = mesh.sort('mass', reverse=False)

    # make sure we have all the columns
    assert all(col in mesh2 for col in mesh)

    mass = numpy.concatenate(comm.allgather(mesh['mass']))
    sorted_mass = numpy.concatenate(comm.allgather(mesh2['mass']))
    assert_array_equal(numpy.sort(mass), sorted_mass)


@MPITest([4])
def test_tsc_interlacing(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-2, BoxSize=512., seed=42)

    # interlacing with TSC
    mesh = source.to_mesh(window='tsc', Nmesh=64, interlaced=True, compensated=True)

    # compute the power spectrum -- should be flat shot noise
    # if the compensation worked
    r = FFTPower(mesh, mode='1d', kmin=0.02)

@MPITest([1])
def test_paint_chunksize(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-2, BoxSize=512., seed=42)

    # interlacing with TSC
    mesh = source.to_mesh(window='tsc', Nmesh=64, interlaced=True, compensated=True)

    with set_options(paint_chunk_size=source.csize // 4):
        r1 = mesh.paint()

    with set_options(paint_chunk_size=source.csize):
        r2 = mesh.paint()

    assert_allclose(r1, r2)

@MPITest([4])
def test_cic_interlacing(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-2, BoxSize=512., seed=42)

    # interlacing with TSC
    mesh = source.to_mesh(window='cic', Nmesh=64, interlaced=True, compensated=True)

    # compute the power spectrum -- should be flat shot noise
    # if the compensation worked
    r = FFTPower(mesh, mode='1d', kmin=0.02)

@MPITest([4])
def test_setters(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-2, BoxSize=512., seed=42)

    # make the mesh
    mesh = source.to_mesh(window='cic', Nmesh=64, interlaced=True, compensated=True)

    assert mesh.compensated == True
    mesh.compensated = False
    assert mesh.compensated == False

    assert mesh.interlaced == True
    mesh.interlaced = False
    assert mesh.interlaced == False

    assert mesh.window == 'cic'
    mesh.window = 'tsc'
    assert mesh.window == 'tsc'

@MPITest([4])
def test_bad_window(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-2, BoxSize=512., seed=42)

    # make the mesh
    mesh = source.to_mesh(window='cic', Nmesh=64, interlaced=True, compensated=True)

    # no such window
    with pytest.raises(Exception):
        mesh.window = "BAD"

@MPITest([4])
def test_no_compensation(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=3e-2, BoxSize=512., seed=42)

    # make the mesh
    mesh = source.to_mesh(window='cic', Nmesh=64, interlaced=True, compensated=True)

    # no compensation for this window
    mesh.window = 'db6'

    # cannot compute compensation
    with pytest.raises(ValueError):
        actions = mesh.actions

@MPITest([1, 4])
def test_copy(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=0.2e-3, BoxSize=1024., seed=42)
    source['TEST'] = 10
    source.attrs['TEST'] = 'TEST'

    # make the mesh
    source = source.to_mesh(Nmesh=32)

    # store original data
    data = {}
    for col in source:
        data[col] = source[col].compute()

    # make copy
    copy = source.copy()

    # modify original
    source['Position'] += 100.
    source['Velocity'] *= 10.

    # check data is equal to original
    for col in copy:
        assert_array_equal(copy[col].compute(), data[col])

    # check meta-data
    for k in source.attrs:
        assert k in copy.attrs

@MPITest([4])
def test_slice(comm):
    CurrentMPIComm.set(comm)

    # the CatalogSource
    source = UniformCatalog(nbar=0.2e-3, BoxSize=1024., seed=42)
    source['TEST'] = 10.

    # the mesh
    source = source.to_mesh(Nmesh=32)

    # slice a subset
    subset = source[:10]
    assert all(col in subset for col in source.columns)
    assert isinstance(subset, source.__class__)
    assert len(subset) == 10
    assert_array_equal(subset['Position'], source['Position'].compute()[:10])

    subset = source[[0,1,2]]
    assert_array_equal(subset['Position'], source['Position'].compute()[[0,1,2]])

@MPITest([4])
def test_view(comm):
    CurrentMPIComm.set(comm)

    # the CatalogSource
    source = UniformCatalog(nbar=0.2e-3, BoxSize=1024., seed=42)
    source['TEST'] = 10.
    source.attrs['TEST'] = 10.0

    # the mesh
    mesh = source.to_mesh(Nmesh=32)

    # view
    view = mesh.view()
    assert view.base is mesh
    assert isinstance(view, mesh.__class__)

    # check meta-data
    for k in mesh.attrs:
        assert k in view.attrs

    # adding columns to the view changes original source
    view['TEST2'] = 5.0
    assert 'TEST2' in source



from nbodykit.cosmology import CorrelationFunction
from nbodykit.cosmology import Cosmology, xi_to_pk, pk_to_xi
from nbodykit.cosmology import LinearPower, HalofitPower, ZeldovichPower
import numpy
from numpy.testing import assert_allclose

def test_mcfit():

    c = Cosmology()
    Plin = LinearPower(c, redshift=0)

    # do Pk to CF
    k = numpy.logspace(-4, 2, 1024)
    CF = pk_to_xi(k, Plin(k))

    # do CF to Pk
    r = numpy.logspace(-3, 3, 1024)
    Pk2 = xi_to_pk(r, CF(r))(k)

    idx = (k>1e-2)&(k<10.)
    assert_allclose(Pk2[idx], Plin(k[idx]), rtol=1e-2)

def test_linear():

    # linear power
    c = Cosmology()
    Plin = LinearPower(c, redshift=0)

    # desired separation (in Mpc/h)
    r = numpy.logspace(0, numpy.log10(150), 500)

    # linear correlation
    CF = CorrelationFunction(Plin)
    CF.sigma8 = 0.82
    xi1 = CF(100.)

    assert_allclose(CF.redshift, CF.attrs['redshift'])
    assert_allclose(CF.sigma8, CF.attrs['sigma8'])

    # change sigma8
    CF.sigma8 = 0.75
    xi2 = CF(100.)
    assert_allclose(xi1/xi2, (0.82/0.75)**2, rtol=1e-2)

    # change redshift
    CF.redshift = 0.55
    xi3 = CF(100.)
    D2 = CF.cosmo.scale_independent_growth_factor(0.)
    D3 = c.scale_independent_growth_factor(0.55)
    assert_allclose(xi2.max()/xi3.max(), (D2/D3)**2, rtol=1e-2)



def test_halofit():

    # nonlinear power
    Pnl = HalofitPower(Cosmology(), redshift=0)

    # desired separation (in Mpc/h)
    r = numpy.logspace(0, numpy.log10(150), 500)

    # nonlinear correlation
    CF = CorrelationFunction(Pnl)

    xi = CF(r)

def test_zeldovich():

    # zeldovich power
    Pzel = ZeldovichPower(Cosmology(), redshift=0)

    # desired separation (in Mpc/h)
    r = numpy.logspace(0, numpy.log10(150), 500)

    # zeldovich correlation
    CF = CorrelationFunction(Pzel)

    xi = CF(r)



from nbodykit import CurrentMPIComm
from nbodykit.base.catalog import CatalogSource
from nbodykit.source import ArrayCatalog

from nbodykit.transform import SkyToUnitSphere
import numpy
import logging

class FiberCollisions(object):
    """
    Run an angular FOF algorithm to determine fiber collision
    groups from an input catalog, and then assign fibers such that
    the maximum amount of object receive a fiber.

    This amounts to determining the following population of objects:

    - population 1:
        the maximal "clean" sample of objects in which each object is not
        angularly collided with any other object in this subsample
    - population 2:
        the potentially-collided objects; these objects are those
        that are fiber collided + those that have been "resolved"
        due to multiple coverage in tile overlap regions

    Results are computed when the object is inititalized. See the documenation
    of :func:`~FiberCollisions.run` for the attributes storing the results.

    Parameters
    ----------
    ra : array_like
        the right ascension coordinate column
    dec : array_like
        the declination coordinate column
    collision_radius : float, optional
        the size of the angular collision radius (in degrees); default
        is 62 arcseconds
    seed : int, optional
        the random seed to use when determining which objects get fibers
    degrees : bool, optional
        set to `True` if the units of ``ra`` and ``dec`` are degrees

    References
    ----------
    - `Guo et al. 2010 <http://arxiv.org/abs/1111.6598>`_
    """
    logger = logging.getLogger('FiberCollisions')

    @CurrentMPIComm.enable
    def __init__(self, ra, dec, collision_radius=62/60./60., seed=None,
                    degrees=True, comm=None):

        # compute the pos
        ra = CatalogSource.make_column(ra)
        dec = CatalogSource.make_column(dec)
        pos = SkyToUnitSphere(ra, dec, degrees=degrees).compute()

        # make the source
        dt = numpy.dtype([('Position', (pos.dtype.str, 3))])
        pos = numpy.squeeze(pos.view(dtype=dt))
        source = ArrayCatalog(pos, BoxSize=numpy.array([2., 2., 2.]))

        self.source = source
        self.comm = source.comm

        # set the seed randomly if it is None
        if seed is None:
            if self.comm.rank == 0:
                seed = numpy.random.randint(0, 4294967295)
            seed = self.comm.bcast(seed)

        # save the attrs
        self.attrs = {}
        self.attrs['collision_radius'] = collision_radius
        self.attrs['seed'] = seed
        self.attrs['degrees'] = degrees

        # store collision radius in radians
        self._collision_radius_rad = numpy.deg2rad(collision_radius)
        if self.comm.rank == 0:
            self.logger.info("collision radius in degrees = %.4f" %collision_radius)

        # compute
        self.run()

    def run(self):
        """
        Run the fiber assignment algorithm. This attaches the following
        attribute to the object:

        - :attr:`labels`

        .. note::

            The :attr:`labels` attribute has a 1-to-1 correspondence with
            the rows in the input source.

        Attributes
        ----------
        labels: :class:`~nbodykit.source.catalog.array.ArrayCatalog`; size: :attr:`size`
            a CatalogSource that has the following columns:

            - Label :
                the group labels for each object in the input
                CatalogSource; label == 0 objects are not in a group
            - Collided :
                a flag array specifying which objects are collided, i.e., do
                not receive a fiber
            - NeighborID :
                for those objects that are collided, this gives the (global)
                index of the nearest neighbor on the sky (0-indexed) in
                the input catalog ``source``, else it is set to -1
        """
        from astropy.utils.misc import NumpyRNGContext
        from nbodykit.algorithms import FOF

        # angular FOF: labels gives the global group ID corresponding to each
        # object in Position on this rank
        fof = FOF(self.source, self._collision_radius_rad, 1, absolute=True)

        # assign the fibers (in parallel)
        with NumpyRNGContext(self.attrs['seed']):
            collided, neighbors = self._assign_fibers(fof.labels)

        # all reduce to get summary statistics
        N_pop1 = self.comm.allreduce((collided^1).sum())
        N_pop2 = self.comm.allreduce((collided).sum())
        f = N_pop2 * 1. / (N_pop1 + N_pop2)

        # print out some info
        if self.comm.rank == 0:

            self.logger.info("population 1 (clean) size = %d" %N_pop1)
            self.logger.info("population 2 (collided) size = %d" %N_pop2)
            self.logger.info("collision fraction = %.4f" %f)

        # return a structured array
        d = list(zip(['Label', 'Collided', 'NeighborID'], [fof.labels, collided, neighbors]))
        dtype = numpy.dtype([(col, x.dtype) for col, x in d])
        result = numpy.empty(len(fof.labels), dtype=dtype)
        for col, x in d: result[col] = x

        # make a particle source
        self.labels = ArrayCatalog(result, comm=self.comm, **self.source.attrs)

    def _assign_fibers(self, Label):
        """
        Internal function to divide the data by collision group
        across ranks and assign fibers, such that the minimum
        number of objects are collided out of the survey
        """
        import mpsort
        from mpi4py import MPI

        mask = Label != 0
        dtype = numpy.dtype([
                ('Position', ('f4', 3)),
                ('Label', ('i4')),
                ('Rank', ('i4')),
                ('Index', ('i4')),
                ('Collided', ('i4')),
                ('NeighborID', ('i4'))
                ])
        PIG = numpy.empty(mask.sum(), dtype=dtype)
        PIG['Label'] = Label[mask]
        size = len(Label)
        size = self.comm.allgather(size)
        Ntot = sum(size)
        offset = sum(size[:self.comm.rank])
        PIG['Index'] = offset + numpy.where(mask == True)[0]
        del Label

        Position = self.source.compute(self.source['Position'])
        PIG['Position'] = Position[mask]
        del Position
        Ntot = self.comm.allreduce(len(mask))
        Nhalo = self.comm.allreduce(
            PIG['Label'].max() if len(PIG['Label']) > 0 else 0, op=MPI.MAX) + 1

        # now count number of particles per halo
        PIG['Rank'] = PIG['Label'] % self.comm.size
        cnt = numpy.bincount(PIG['Rank'], minlength=self.comm.size)
        Nlocal = self.comm.allreduce(cnt)[self.comm.rank]

        # sort by rank and then label
        PIG2 = numpy.empty(Nlocal, PIG.dtype)
        mpsort.sort(PIG, orderby='Rank', out=PIG2, comm=self.comm)
        assert (PIG2['Rank'] == self.comm.rank).all()
        PIG2.sort(order=['Label'])

        if self.comm.rank == 0:
            self.logger.info('total number of collision groups = %d', Nhalo-1)
            self.logger.info("Started fiber assignment")

        # loop over unique group ids
        for group_id in numpy.unique(PIG2['Label']):
            start = PIG2['Label'].searchsorted(group_id, side='left')
            end = PIG2['Label'].searchsorted(group_id, side='right')
            N = end-start
            assert(PIG2['Label'][start:end] == group_id).all()

            # pairs (random selection)
            if N == 2:

                # randomly choose, with fixed local seed
                which = numpy.random.choice([0,1])

                indices = [start+which, start+(which^1)]
                PIG2['Collided'][indices] = [1, 0]
                PIG2['NeighborID'][indices] = [PIG2['Index'][start+(which^1)], -1]
            # multiplets (minimize collidedness)
            elif N > 2:

                collided, nearest = self._assign_multiplets(PIG2['Position'][start:end])
                PIG2['Collided'][start:end] = collided[:]
                PIG2['NeighborID'][start:end] = -1
                if len(nearest):
                    PIG2['NeighborID'][start:end][collided] = PIG2['Index'][start+nearest][:]

        if self.comm.rank == 0: self.logger.info("Finished fiber assignment")

        # return to the order specified by the global unique index
        mpsort.sort(PIG2, orderby='Index', out=PIG, comm=self.comm)

        # return arrays including the objects not in any groups
        collided = numpy.zeros(size[self.comm.rank], dtype='i4')
        collided[mask] = PIG['Collided'][:]
        neighbors = numpy.zeros(size[self.comm.rank], dtype='i4') - 1
        neighbors[mask] = PIG['NeighborID'][:]

        del PIG
        return collided, neighbors

    def _assign_multiplets(self, Position):
        """
        Internal function to assign the maximal amount of fibers
        in collision groups of size N > 2
        """
        from scipy.spatial.distance import pdist, squareform

        def count(slice, n):
            return n[numpy.nonzero(slice)[0]].sum()

        # first shuffle the member ids, so we select random element when tied
        N = len(Position)
        group_ids = list(range(N))

        collided_ids = []
        while len(group_ids) > 1:

            # compute dists and find where dists < collision radius
            dists = squareform(pdist(Position[group_ids], metric='euclidean'))
            numpy.fill_diagonal(dists, numpy.inf) # ignore self-pairs
            collisions = dists <= self._collision_radius_rad

            # total # of collisions for each group member
            n_collisions = numpy.sum(collisions, axis=0)

            # total # of collisions for those objects that collide with each group member
            n_other = numpy.apply_along_axis(count, 0, collisions, n_collisions)

            # remove object that has most # of collisions
            # and those colliding objects have least # of collisions
            idx = numpy.where(n_collisions == n_collisions.max())[0]

            # choose randomly, with a fixed local seed
            ii = numpy.random.choice(numpy.where(n_other[idx] == n_other[idx].min())[0])
            collided_index = idx[ii]

            # make the collided galaxy and remove from group
            collided_id = group_ids.pop(collided_index)

            # only make this a collided object if its n_collisions > 0
            # if n_collisions = 0, then the object can get a fiber for free
            if n_collisions[collided_index] > 0:
                collided_ids.append(collided_id)

        # compute the nearest neighbors
        neighbor_ids = []
        group_indices = list(range(N))
        dists = squareform(pdist(Position, metric='euclidean'))
        uncollided = [i for i in group_indices if i not in collided_ids]

        for i in sorted(collided_ids):
            neighbor = uncollided[dists[i][uncollided].argmin()]
            neighbor_ids.append(neighbor)

        collided = numpy.zeros(N, dtype=bool)
        collided[collided_ids] = True
        return collided, neighbor_ids
